/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sp_rpc.h"
#include "sp_rpc_internal.h"

#include <string.h>

/* Prototypes: */
static void sp_rpc_cb_free (SPRpcCb * data);

/* MAIN FUNCTIONS ***********************************************************/

/**
 * sp_rpc_new:
 * @returns: a new SPRpc with a reference count of 1.
 *
 * Creates a new SPRpc that must be initialized before using though the 
 * sp_rpc_set_mailbox() and the spread functions (SP_connect(),
 * SP_join(), ...).
 *
 * Use sp_rpc_unref() to free this data.
 */
SPRpc *
sp_rpc_new (void)
{
  SPRpc *data;

  data = g_malloc0 (sizeof (SPRpc));
  sp_rpc_ref (data);

  return data;
}

/**
 * sp_rpc_ref:
 * @rpc: a SPRpc
 *
 * Increases the reference count of a SPRpc. 
 */
void
sp_rpc_ref (SPRpc * rpc)
{
  g_return_if_fail (rpc != NULL);

  rpc->ref++;
}

/**
 * sp_rpc_unref:
 * @rpc: a SPRpc
 *
 * Decreases the reference count of a SPRpc. When its reference count drops
 * to 0, the SPRpc is finalized.
 */
void
sp_rpc_unref (SPRpc * rpc)
{
  g_return_if_fail (rpc != NULL);

  rpc->ref--;

  if (rpc->ref > 0)
    return;

  if (rpc->group)
    g_free (rpc->group);

  if (rpc->id)
    g_free (rpc->id);

  if (rpc->methods)
    {
      g_list_foreach (rpc->methods, (GFunc) sp_rpc_cb_free, NULL);
      g_list_free (rpc->methods);
    }

  if (rpc->responses)
    {
      g_list_foreach (rpc->responses, (GFunc) sp_rpc_cb_free, NULL);
      g_list_free (rpc->responses);
    }

  g_free (rpc);
}

/* MAILBOX FUNCTIONS ********************************************************/

/**
 * sp_rpc_get_mailbox:
 * @rpc: a SPRpc
 * @returns: the mailbox of the SPRpc object
 *
 * Returns the internal mailbox element of a SPRpc. This mailbox can be used
 * for the initialization of spread through (SP_connect(), SP_join(), ...) or
 * for the disconnect procedure (SP_disconnect(), ...).
 */
mailbox
sp_rpc_get_mailbox (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, 0);

  return rpc->mailbox;
}

/**
 * sp_rpc_set_mailbox:
 * @rpc: a SPRpc
 * @mailbox: a Spread mailbox, or 0
 *
 * Sets a new mailbox for a SPRpc. Normally, this function is used with a new
 * mailbox could be connected to the spread ring.
 */
void
sp_rpc_set_mailbox (SPRpc * rpc, mailbox mailbox)
{
  g_return_if_fail (rpc != NULL);

  rpc->mailbox = mailbox;
}

/* GROUP FUNCTIONS ********************************************************/

/**
 * sp_rpc_get_group:
 * @rpc: a SPRpc
 * @returns: The group of a SPRpc. This string is owned by the SPRpc and must
 * not be modified or freed. 
 *
 * Returns the group string of a SPRpc. 
 */
const gchar *
sp_rpc_get_group (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, 0);

  return rpc->group;
}

/**
 * sp_rpc_set_group:
 * @rpc: a SPRpc
 * @group: the group
 *
 * Sets a new group for a SPRpc.
 */
void
sp_rpc_set_group (SPRpc * rpc, gchar * group)
{
  g_return_if_fail (rpc != NULL);
  g_return_if_fail (group != NULL);

  if (rpc->group)
    g_free (rpc->group);

  rpc->group = g_strdup (group);
}

/**
 * sp_rpc_group_exists:
 * @rpc: a SPRpc
 * @returns: TRUE/FALSE
 *
 * Returns TRUE if the group exists into the SPRpc object
 */
gboolean
sp_rpc_group_exists (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, FALSE);

  return rpc->group ? TRUE : FALSE;
}

/* ID FUNCTIONS ********************************************************/

/**
 * sp_rpc_get_id:
 * @rpc: a SPRpc
 * @returns: The id of a SPRpc. This string is owned by the SPRpc and must
 * not be modified or freed. 
 *
 * Returns the id string of a SPRpc. 
 */
const gchar *
sp_rpc_get_id (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, 0);

  return rpc->id;
}

/**
 * sp_rpc_set_id:
 * @rpc: a SPRpc
 * @id: the id
 *
 * Sets a new id for a SPRpc.
 */
void
sp_rpc_set_id (SPRpc * rpc, gchar * id)
{
  g_return_if_fail (rpc != NULL);
  g_return_if_fail (id != NULL);

  if (rpc->id)
    g_free (rpc->id);

  rpc->id = g_strdup (id);
}

/**
 * sp_rpc_id_exists:
 * @rpc: a SPRpc
 * @returns: TRUE/FALSE
 *
 * Returns TRUE if the id exists into the SPRpc object
 */
gboolean
sp_rpc_id_exists (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, FALSE);

  return rpc->id ? TRUE : FALSE;
}

/* (UN)REGISTER FUNCTIONS ***************************************************/

/**
 * sp_rpc_register:
 * @rpc: a SPRpc
 * @method_name: the name of the method
 * @method_description: a description of the method or NULL
 * @params: a SPRpcParams with the SPRpcValues for this method
 * @method_function: a pointer to a SPRpcMethod
 * @user_data: data to pass to the function calls
 * @returns: a SPRpcMethodID of the new registerd method
 *
 * Adds a new method to a SPRpc. The new method will be call when a
 * request is received by the spread ring.
 *
 * The params describes the inputs of the method. If you specify the type of
 * the input, they will be validated when a client calls this method. If you
 * specify also a value, if the client doesn't send one of them, it will be
 * created with the value you specified here. Use SP_RPC_VALUE_NULL if you
 * want to accept any kind of value.
 *
 * If a method with this name already exists, it will be replaced with this
 * new one.
 */
SPRpcMethodID
sp_rpc_register (SPRpc * rpc, const gchar * method_name,
		 const gchar * method_description, SPRpcParams * params,
		 SPRpcMethod method_function, gpointer user_data)
{
  SPRpcCb *method;
  GList *list;

  g_return_val_if_fail (rpc != NULL, 0);
  g_return_val_if_fail (method_name != NULL, 0);
  g_return_val_if_fail (method_function != NULL, 0);

  for (list = rpc->methods; list; list = list->next)
    {
      method = list->data;

      if (!g_utf8_collate (method_name, method->name))
	{
	  sp_rpc_unregister (rpc, method->number);
	  break;
	}
    }

  method = g_malloc0 (sizeof (SPRpcCb));

  if (params)
    {
      sp_rpc_params_ref (params);
      method->params = params;
    }

  method->name = g_strdup (method_name);

  if (method_description)
    method->description = g_strdup (method_description);

  method->method_function = method_function;
  method->user_data = user_data;

  method->number = ++rpc->method_last_numb;
  rpc->methods = g_list_prepend (rpc->methods, method);

  return method->number;
}

/**
 * sp_rpc_unregister:
 * @rpc: a SPRpc
 * @method_id: the ID of the method
 *
 * Removes a SPRpcMethod from the SPRpc.
 */
void
sp_rpc_unregister (SPRpc * rpc, SPRpcMethodID method_id)
{
  GList *list;

  g_return_if_fail (rpc != NULL);
  g_return_if_fail (method_id != 0);

  for (list = rpc->methods; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (data->number == method_id)
	{
	  rpc->methods = g_list_remove (rpc->methods, data);
	  sp_rpc_cb_free (data);
	  return;
	}
    }

  g_return_if_fail (sp_rpc_is_registered (rpc, method_id) != TRUE);
}

/**
 * sp_rpc_is_registered:
 * @rpc: a SPRpc
 * @method_id: the ID of the method
 * @returns: TRUE or FALSE
 *
 * Returns TRUE if the method ID is valid and registerd into a SPRpc.
 */
gboolean
sp_rpc_is_registered (SPRpc * rpc, SPRpcMethodID method_id)
{
  GList *list;

  g_return_val_if_fail (rpc != NULL, FALSE);
  g_return_val_if_fail (method_id != 0, FALSE);

  for (list = rpc->methods; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (data->number == method_id)
	return TRUE;
    }

  return FALSE;
}

/* (UN)REGISTER CALL FUNCTIONS **********************************************/
static gint
sp_rpc_send (SPRpc * rpc, gchar * message, gsize length)
{
  gint ret;

  ret =
    SP_multicast (rpc->mailbox, FIFO_MESS, rpc->group, 1, length, message);

  if (ret <= 0)
    {
      switch (ret)
	{
	case ILLEGAL_SESSION:
	  return SP_RPC_STATUS_ILLEGAL_SESSION;

	case ILLEGAL_MESSAGE:
	  return SP_RPC_STATUS_ILLEGAL_MESSAGE;

	case CONNECTION_CLOSED:
	  return SP_RPC_STATUS_CONNECTION_CLOSED;
	}

      return SP_RPC_STATUS_UNKNOWN_ERROR;
    }

  return SP_RPC_STATUS_DONE;
}

/**
 * sp_rpc_call:
 * @rpc: a SPRpc
 * @method_name: the method name
 * @params: the params to send
 * @request_type: the type of the request
 * @response_id: the location of a ID, or NULL
 * @response_function: the pointer of the callback, or NULL
 * @user_data: data to pass to the function calls
 * @returns: a status about the message
 *
 * Sends the request for the remote procedure call. The params must be compatible
 * with the called method.
 *
 * If @response_function is not NULL, it will be called with params
 * when the remote method returns them.
 */
SPRpcStatus
sp_rpc_call (SPRpc * rpc, const gchar * method_name, SPRpcParams * params,
	     SPRpcReqType request_type, SPRpcResponseID * response_id,
	     SPRpcResponse response_function, gpointer user_data)
{
  SPRpcCb *cb;
  gchar *message;
  guint length;
  gint ret;

  gint progress_id;

  g_return_val_if_fail (rpc != NULL, 0);
  g_return_val_if_fail (method_name != NULL, 0);

  g_return_val_if_fail (sp_rpc_group_exists (rpc) == TRUE, 0);
  g_return_val_if_fail (sp_rpc_id_exists (rpc) == TRUE, 0);

  progress_id = rpc->response_last_numb + 1;

  switch (request_type)
    {
    case SP_RPC_REQ_TYPE_CALL:
      message =
	sp_rpc_json_create (rpc->id, progress_id, SP_RPC_JSON_CALL,
			    method_name, params, &length);
      break;

    case SP_RPC_REQ_TYPE_GET_PARAMS:
      message =
	sp_rpc_json_create (rpc->id, progress_id, SP_RPC_JSON_PARAMS,
			    method_name, NULL, &length);
      break;

    case SP_RPC_REQ_TYPE_EXISTS:
      message =
	sp_rpc_json_create (rpc->id, progress_id, SP_RPC_JSON_EXISTS,
			    method_name, NULL, &length);
      break;

    default:
      return SP_RPC_STATUS_ILLEGAL_MESSAGE;
    }

  ret = sp_rpc_send (rpc, message, length);
  g_free (message);

  if (ret != SP_RPC_STATUS_DONE)
    return ret;

  if (!response_function)
    return SP_RPC_STATUS_DONE;

  cb = g_malloc0 (sizeof (SPRpcCb));

  if (params)
    {
      sp_rpc_params_ref (params);
      cb->params = params;
    }

  cb->name = g_strdup (method_name);

  cb->response_function = response_function;
  cb->user_data = user_data;

  cb->number = ++rpc->response_last_numb;
  rpc->responses = g_list_prepend (rpc->responses, cb);

  if (response_id)
    *response_id = cb->number;

  return SP_RPC_STATUS_DONE;
}

/**
 * sp_rpc_call_list:
 * @rpc: a SPRpc
 * @response_id: a location of the response ID, or NULL
 * @response_function: the pointer of the callback
 * @user_data: data to pass to the function calls
 * @returns: the status of the operation
 *
 * Sends the request for the list of the calls
 */
SPRpcStatus
sp_rpc_call_list (SPRpc * rpc, SPRpcResponseID * response_id,
		  SPRpcResponse response_function, gpointer user_data)
{
  guint length;
  gchar *message;
  SPRpcCb *cb;
  SPRpcStatus status;

  guint progress_id;

  g_return_val_if_fail (rpc != NULL, 0);
  g_return_val_if_fail (response_function != NULL, 0);

  progress_id = ++rpc->response_last_numb;
  message =
    sp_rpc_json_create (rpc->id, progress_id, SP_RPC_JSON_LIST, NULL, NULL,
			&length);

  status = sp_rpc_send (rpc, message, length);
  g_free (message);

  if (status != SP_RPC_STATUS_DONE)
    return status;

  cb = g_malloc0 (sizeof (SPRpcCb));

  cb->response_function = response_function;
  cb->user_data = user_data;

  cb->number = progress_id;
  rpc->responses = g_list_prepend (rpc->responses, cb);

  if (response_id)
    *response_id = cb->number;

  return SP_RPC_STATUS_DONE;
}

/**
 * sp_rpc_call_remove:
 * @rpc: a SPRpc
 * @response_id: the ID of the calling operation
 *
 * Removes the calling operation from a SPRpc. Any other value from the
 * request will be ignored.
 */
void
sp_rpc_call_remove (SPRpc * rpc, SPRpcResponseID response_id)
{
  GList *list;

  g_return_if_fail (rpc != NULL);
  g_return_if_fail (response_id != 0);

  for (list = rpc->responses; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (data->number == response_id)
	{
	  rpc->responses = g_list_remove (rpc->responses, data);
	  sp_rpc_cb_free (data);
	  return;
	}
    }

  g_return_if_fail (sp_rpc_is_registered (rpc, response_id) != TRUE);
}

/**
 * sp_rpc_call_is_registered:
 * @rpc: a SPRpc
 * @response_id: the ID of the calling operation
 * @returns: TRUE/FALSE
 *
 * Returns TRUE if the response_id is valid and registered to the SPRpc.
 */
gboolean
sp_rpc_call_is_registered (SPRpc * rpc, SPRpcResponseID response_id)
{
  GList *list;

  g_return_val_if_fail (rpc != NULL, FALSE);
  g_return_val_if_fail (response_id != 0, FALSE);

  for (list = rpc->responses; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (data->number == response_id)
	return TRUE;
    }

  return FALSE;
}

/* LOOP FUNCTIONS ***********************************************************/

/**
 * sp_rpc_run:
 * @rpc: a SPRpc
 * @returns: a status
 *
 * Execs the SPRpc loop for the messages. It will concluded by a sp_rpc_quit.
 */
SPRpcStatus
sp_rpc_run (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, 0);

  g_return_val_if_fail (sp_rpc_group_exists (rpc) == TRUE, 0);
  g_return_val_if_fail (sp_rpc_id_exists (rpc) == TRUE, 0);

  rpc->is_running = TRUE;

  while (rpc->is_running)
    {
      SPRpcStatus status;
      gboolean p;

      if ((status = sp_rpc_event_pending (rpc, &p)) != SP_RPC_STATUS_DONE)
	return status;

      if (p == FALSE)
	{
	  g_usleep (SP_RPC_RUN_TIMEOUT);
	  continue;
	}

      if ((status = sp_rpc_iteration (rpc, TRUE)) != SP_RPC_STATUS_DONE)
	return status;
    }

  return SP_RPC_STATUS_DONE;
}

/**
 * sp_rpc_is_running:
 * @rpc: a SPRpc
 * @returns: TRUE/FALSE
 *
 * Returns TRUE isthe SPRpc loop is running.
 */
gboolean
sp_rpc_is_running (SPRpc * rpc)
{
  g_return_val_if_fail (rpc != NULL, FALSE);
  return rpc->is_running;
}

/**
 * sp_rpc_quit:
 * @rpc: a SPRpc
 *
 * Quits from the SPRpc loop.
 */
void
sp_rpc_quit (SPRpc * rpc)
{
  g_return_if_fail (rpc != NULL);
  rpc->is_running = FALSE;
}

static SPRpcStatus sp_rpc_iteration_parse (SPRpc * rpc, gchar * buffer,
					   gsize size);

/**
 * sp_rpc_iteration:
 * @rpc: a SPRpc
 * @may_block: whether the call may block
 * @returns: a status of the operation
 *
 * Runs a single iteration for the given SPRpc.
 */
SPRpcStatus
sp_rpc_iteration (SPRpc * rpc, gboolean may_block)
{
  service service;
  gchar sender[MAX_GROUP_NAME];

  gint num_groups;
  gint ret;

#define SP_RPC_RECEIVE_MAX_GROUP	64
  gchar groups_real[SP_RPC_RECEIVE_MAX_GROUP][MAX_GROUP_NAME];
  gboolean groups_allocated;
  gsize groups_size;
  gpointer groups;

  gint16 mess_type;
  gint endian_mismatch;

  gboolean buffer_allocated;
  gchar buffer_real[1024];
  gsize buffer_size;
  gchar *buffer;

  SPRpcStatus status;

  g_return_val_if_fail (rpc != NULL, 0);

  g_return_val_if_fail (sp_rpc_group_exists (rpc) == TRUE, 0);
  g_return_val_if_fail (sp_rpc_id_exists (rpc) == TRUE, 0);

  if (may_block == FALSE)
    {
      gboolean p;

      if ((status = sp_rpc_event_pending (rpc, &p)) != SP_RPC_STATUS_DONE)
	return status;

      if (p == FALSE)
	return SP_RPC_STATUS_DONE;
    }

  service = 0;

  buffer = buffer_real;
  buffer_allocated = FALSE;
  buffer_size = sizeof (buffer_real);

  groups = groups_real;
  groups_allocated = FALSE;
  groups_size = SP_RPC_RECEIVE_MAX_GROUP;

  while (1)
    {
      ret =
	SP_receive (rpc->mailbox, &service, sender, groups_size, &num_groups,
		    groups, &mess_type, &endian_mismatch, buffer_size,
		    buffer);

      if (ret > 0)
	break;

      if (ret == ILLEGAL_SESSION || ret == CONNECTION_CLOSED
	  || ret == ILLEGAL_MESSAGE)
	{
	  if (groups_allocated)
	    g_free (groups);

	  if (buffer_allocated)
	    g_free (buffer);

	  switch (ret)
	    {
	    case ILLEGAL_SESSION:
	      return SP_RPC_STATUS_ILLEGAL_SESSION;

	    case ILLEGAL_MESSAGE:
	      return SP_RPC_STATUS_ILLEGAL_MESSAGE;

	    case CONNECTION_CLOSED:
	      return SP_RPC_STATUS_CONNECTION_CLOSED;
	    }
	}

      if (ret == GROUPS_TOO_SHORT)
	{
	  gint i;
	  gchar **g;

	  if (groups_allocated)
	    g_strfreev (groups);

	  groups_allocated = TRUE;
	  groups_size += SP_RPC_RECEIVE_MAX_GROUP;
	  g = g_malloc (sizeof (gchar *) * (groups_size + 1));

	  for (i = 0; i < groups_size; i++)
	    g[i] = g_malloc (MAX_GROUP_NAME);

	  g[i] = NULL;
	  groups = g;

	  continue;
	}

      if (ret == BUFFER_TOO_SHORT)
	{
	  buffer_size += sizeof (buffer_real);

	  if (!buffer_allocated)
	    {
	      buffer_allocated = TRUE;
	      buffer = g_malloc (buffer_size);
	    }

	  else
	    buffer = g_realloc (buffer, buffer_size);

	  continue;
	}

      return SP_RPC_STATUS_UNKNOWN_ERROR;
    }

  status = sp_rpc_iteration_parse (rpc, buffer, ret);

  if (groups_allocated)
    g_strfreev (groups);

  if (buffer_allocated)
    g_free (buffer);

  return status;
}

static SPRpcStatus sp_rpc_iteration_call (SPRpc * rpc, gchar * id,
					  guint progress_id, gchar * method,
					  SPRpcParams * params);
static SPRpcStatus sp_rpc_iteration_list (SPRpc * rpc, gchar * id,
					  guint progress_id);
static SPRpcStatus sp_rpc_iteration_exists (SPRpc * rpc, gchar * id,
					    guint progress_id,
					    gchar * method);
static SPRpcStatus sp_rpc_iteration_params (SPRpc * rpc, gchar * id,
					    guint progress_id,
					    gchar * method);
static SPRpcStatus sp_rpc_iteration_response (SPRpc * rpc, gchar * id,
					      guint progress_id,
					      SPRpcParams * params);

static SPRpcStatus
sp_rpc_iteration_parse (SPRpc * rpc, gchar * buffer, gsize size)
{
  SPRpcParams *params;
  SPRpcJsonType type;
  gchar *method;

  guint progress_id;
  gchar *id;

  SPRpcStatus status = SP_RPC_STATUS_DONE;

  if (sp_rpc_json_parse
      (buffer, size, &id, &progress_id, &type, &method, &params) == FALSE)
    return SP_RPC_STATUS_DONE;

  switch (type)
    {
    case SP_RPC_JSON_CALL:
      if (!method)
	break;

      status = sp_rpc_iteration_call (rpc, id, progress_id, method, params);
      break;

    case SP_RPC_JSON_LIST:
      status = sp_rpc_iteration_list (rpc, id, progress_id);
      break;

    case SP_RPC_JSON_EXISTS:
      if (!method)
	break;

      status = sp_rpc_iteration_exists (rpc, id, progress_id, method);
      break;

    case SP_RPC_JSON_PARAMS:
      if (!method)
	break;

      status = sp_rpc_iteration_params (rpc, id, progress_id, method);
      break;


    case SP_RPC_JSON_RESPONSE:
      if (!params)
	break;

      status = sp_rpc_iteration_response (rpc, id, progress_id, params);
      break;

    default:
      break;
    }

  if (id)
    g_free (id);

  if (method)
    g_free (method);

  if (params)
    sp_rpc_params_unref (params);

  return status;
}

static SPRpcStatus
sp_rpc_iteration_call (SPRpc * rpc, gchar * id, guint progress_id,
		       gchar * method, SPRpcParams * params)
{
  GList *list;

  for (list = rpc->methods; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (!g_utf8_collate (data->name, method))
	{
	  SPRpcParams *rparams;
	  SPRpcParams *nparams;
	  gchar *message;
	  gsize size;

	  SPRpcStatus status = SP_RPC_STATUS_DONE;

	  nparams = sp_rpc_params_validate (data->params, params);
	  rparams = sp_rpc_params_new ();

	  data->method_function (rpc, data->number, nparams, rparams,
				 data->user_data);

	  if (sp_rpc_params_is_empty (rparams) == FALSE)
	    {
	      message =
		sp_rpc_json_create (id, progress_id, SP_RPC_JSON_RESPONSE,
				    NULL, rparams, &size);

	      status = sp_rpc_send (rpc, message, size);
	      g_free (message);
	    }

	  sp_rpc_params_unref (rparams);

	  if (nparams)
	    sp_rpc_params_unref (nparams);

	  return status;
	}
    }

  return SP_RPC_STATUS_DONE;
}

static SPRpcStatus
sp_rpc_iteration_list (SPRpc * rpc, gchar * id, guint progress_id)
{
  gint ret;
  gsize size;
  gchar *message;
  SPRpcParams *params;

  if (!rpc->methods
      || !(params = sp_rpc_params_new_for_call_list (rpc->methods)))
    return SP_RPC_STATUS_DONE;

  message =
    sp_rpc_json_create (id, progress_id, SP_RPC_JSON_RESPONSE, NULL, params,
			&size);

  ret = sp_rpc_send (rpc, message, size);
  g_free (message);

  sp_rpc_params_unref (params);
  return ret;
}

static SPRpcStatus
sp_rpc_iteration_exists (SPRpc * rpc, gchar * id, guint progress_id,
			 gchar * method)
{
  GList *list;

  for (list = rpc->methods; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (!g_utf8_collate (data->name, method))
	{
	  SPRpcParams *params;
	  gchar *message;
	  gsize size;

	  SPRpcStatus status;

	  params = sp_rpc_params_new_for_call_exists (method);
	  message =
	    sp_rpc_json_create (id, progress_id, SP_RPC_JSON_RESPONSE, NULL,
				params, &size);

	  status = sp_rpc_send (rpc, message, size);
	  g_free (message);

	  sp_rpc_params_unref (params);
	  return status;
	}
    }

  return SP_RPC_STATUS_DONE;
}

static SPRpcStatus
sp_rpc_iteration_params (SPRpc * rpc, gchar * id, guint progress_id,
			 gchar * method)
{
  GList *list;

  for (list = rpc->methods; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (!g_utf8_collate (data->name, method))
	{
	  gchar *message;
	  gsize size;

	  SPRpcStatus status;

	  message =
	    sp_rpc_json_create (id, progress_id, SP_RPC_JSON_RESPONSE, NULL,
				data->params, &size);

	  status = sp_rpc_send (rpc, message, size);
	  g_free (message);

	  return status;
	}
    }

  return SP_RPC_STATUS_DONE;
}

static SPRpcStatus
sp_rpc_iteration_response (SPRpc * rpc, gchar * id, guint progress_id,
			   SPRpcParams * params)
{
  GList *list;

  if (g_utf8_collate (id, rpc->id))
    return SP_RPC_STATUS_DONE;

  for (list = rpc->responses; list; list = list->next)
    {
      SPRpcCb *data = list->data;

      if (data->number == progress_id)
	{
	  data->response_function (rpc, data->number, params,
				   data->user_data);
	  break;
	}
    }

  return SP_RPC_STATUS_DONE;
}

/**
 * sp_rpc_event_pending
 * @rpc: a SPRpc
 * @pending: location for a boolean
 * @returns: the status
 *
 * Returns if there some events pending
 */
SPRpcStatus
sp_rpc_event_pending (SPRpc * rpc, gboolean * pending)
{
  gint m;

  g_return_val_if_fail (rpc != NULL, 0);
  g_return_val_if_fail (pending != NULL, 0);

  if (!(m = SP_poll (rpc->mailbox)))
    {
      *pending = FALSE;
      return SP_RPC_STATUS_DONE;
    }

  if (m > 0)
    {
      *pending = TRUE;
      return SP_RPC_STATUS_DONE;
    }

  switch (m)
    {
    case ILLEGAL_SESSION:
      return SP_RPC_STATUS_ILLEGAL_SESSION;

    case ILLEGAL_MESSAGE:
      return SP_RPC_STATUS_ILLEGAL_MESSAGE;

    case CONNECTION_CLOSED:
      return SP_RPC_STATUS_CONNECTION_CLOSED;

    default:
      return SP_RPC_STATUS_UNKNOWN_ERROR;
    }
}

/* STATIC INTERNAL FUNCTIONS ************************************************/
static void
sp_rpc_cb_free (SPRpcCb * data)
{
  if (!data)
    return;

  if (data->name)
    g_free (data->name);

  if (data->description)
    g_free (data->description);

  if (data->params)
    sp_rpc_params_unref (data->params);

  g_free (data);
}

/* EOF */
