/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __SP_RPC_H__
#define __SP_RPC_H__

#include <sp.h>
#include <glib.h>

#define	SP_RPC_PROTOCOL_VERSION	"0.1"

extern const gchar *	sp_rpc_version_string;
extern const guint	sp_rpc_version_major;
extern const guint	sp_rpc_version_minor;
extern const guint	sp_rpc_version_micro;

typedef struct sp_rpc_t SPRpc;

typedef struct sp_rpc_value_t SPRpcValue;
typedef struct sp_rpc_params_t SPRpcParams;
typedef GHashTableIter SPRpcIter;

typedef struct sp_rpc_glib_source_t SPRpcGlibSource;

typedef gint SPRpcMethodID;
typedef gint SPRpcResponseID;

typedef enum
{
  SP_RPC_TYPE_NULL,

  SP_RPC_TYPE_NUMBER,
  SP_RPC_TYPE_BOOLEAN,
  
  SP_RPC_TYPE_STRING,

  SP_RPC_TYPE_ARRAY,
  SP_RPC_TYPE_STRUCT
} SPRpcTypeValue;

typedef enum
{
  SP_RPC_STATUS_DONE = 0,
  SP_RPC_STATUS_ILLEGAL_SESSION,
  SP_RPC_STATUS_ILLEGAL_MESSAGE,
  SP_RPC_STATUS_CONNECTION_CLOSED,
  SP_RPC_STATUS_UNKNOWN_ERROR
} SPRpcStatus;

typedef enum
{
  SP_RPC_REQ_TYPE_CALL	= 0,
  SP_RPC_REQ_TYPE_GET_PARAMS,
  SP_RPC_REQ_TYPE_EXISTS
} SPRpcReqType;

typedef gboolean (*SPRpcGlibCallback)	(SPRpcGlibSource * source,
					 SPRpcStatus	status,
					 gpointer	user_data);

typedef void	(*SPRpcMethod)		(SPRpc *	rpc,
					 SPRpcMethodID	method_id,
					 SPRpcParams *	params,
					 SPRpcParams *	return_params,
					 gpointer	user_data);

typedef void	(*SPRpcResponse)	(SPRpc *	rpc,
					 SPRpcResponseID response_id,
					 SPRpcParams *	params,
					 gpointer	user_data);

SPRpc *		sp_rpc_new		(void)
					 G_GNUC_WARN_UNUSED_RESULT;

void		sp_rpc_ref		(SPRpc *	rpc);
void		sp_rpc_unref		(SPRpc *	rpc);

mailbox		sp_rpc_get_mailbox	(SPRpc *	rpc);
void		sp_rpc_set_mailbox	(SPRpc *	rpc,
					 mailbox	mailbox);

const gchar *	sp_rpc_get_group	(SPRpc *	rpc)
					 G_GNUC_WARN_UNUSED_RESULT;
void		sp_rpc_set_group	(SPRpc *	rpc,
					 gchar *	group);
gboolean	sp_rpc_group_exists	(SPRpc *	rpc);

const gchar *	sp_rpc_get_id		(SPRpc *	rpc)
					 G_GNUC_WARN_UNUSED_RESULT;
void		sp_rpc_set_id		(SPRpc *	rpc,
					 gchar *	id);
gboolean	sp_rpc_id_exists	(SPRpc *	rpc);

/* It returns the ID of the new registered method */
SPRpcMethodID	sp_rpc_register		(SPRpc *	rpc,
					 const gchar *	method_name,
					 const gchar *	method_description,
					 SPRpcParams *	params,
					 SPRpcMethod	method_function,
					 gpointer	user_data);
void		sp_rpc_unregister	(SPRpc *	rpc,
					 SPRpcMethodID	method_id);
gboolean	sp_rpc_is_registered	(SPRpc *	rpc,
					 SPRpcMethodID	method_id);

/* Client side: */
SPRpcStatus	sp_rpc_call		(SPRpc *	rpc,
					 const gchar *	method_name,
					 SPRpcParams *	params,
					 SPRpcReqType	request_type,
					 SPRpcResponseID * response_id,
					 SPRpcResponse	response_function,
					 gpointer	user_data);
SPRpcStatus	sp_rpc_call_list	(SPRpc *	rpc,
					 SPRpcResponseID * response_id,
					 SPRpcResponse	response_function,
					 gpointer	user_data);
void		sp_rpc_call_remove	(SPRpc *	rpc,
					 SPRpcResponseID response_id);
gboolean	sp_rpc_call_is_registered
					(SPRpc *	rpc,
					 SPRpcResponseID response_id);

/* Loop: */
SPRpcStatus	sp_rpc_run		(SPRpc *	rpc);
gboolean	sp_rpc_is_running	(SPRpc *	rpc);
void		sp_rpc_quit		(SPRpc *	rpc);

SPRpcStatus	sp_rpc_iteration	(SPRpc *	rpc,
					 gboolean	may_block);
SPRpcStatus	sp_rpc_event_pending	(SPRpc *	rpc,
					 gboolean *	pending);

GSource *	sp_rpc_glib_source	(SPRpc *	rpc)
					 G_GNUC_WARN_UNUSED_RESULT;

/* Values: */
SPRpcValue *	sp_rpc_value_new	(void)
					 G_GNUC_WARN_UNUSED_RESULT;

void		sp_rpc_value_ref	(SPRpcValue *	value);
void		sp_rpc_value_unref	(SPRpcValue *	value);

void		sp_rpc_value_set_type	(SPRpcValue *	value,
					 SPRpcTypeValue	type);
SPRpcTypeValue	sp_rpc_value_get_type	(SPRpcValue *	value);

/* Simple Values: */
gdouble		sp_rpc_value_get_number	(SPRpcValue *	value);
void		sp_rpc_value_set_number	(SPRpcValue *	value,
					 gdouble	v_number);

gboolean	sp_rpc_value_get_boolean(SPRpcValue *	value);
void		sp_rpc_value_set_boolean(SPRpcValue *	value,
					 gboolean	v_boolean);

gchar *		sp_rpc_value_get_string	(SPRpcValue *	value)
					 G_GNUC_WARN_UNUSED_RESULT;
void		sp_rpc_value_set_string	(SPRpcValue *	value,
					 gchar *	v_string);
void		sp_rpc_value_set_string_printf
					(SPRpcValue *	value,
					 gchar *	format,
					 ...);

/* Arrays: */
guint		sp_rpc_value_array_length(SPRpcValue *	varray);

SPRpcValue *	sp_rpc_value_array_get	(SPRpcValue *	varray,
					 guint		id);

gint		sp_rpc_value_array_get_id_by_value
					(SPRpcValue *	varray,
					 SPRpcValue *	v);

void		sp_rpc_value_array_set	(SPRpcValue *	varray,
					 SPRpcValue *	v);

void		sp_rpc_value_array_set_new
					(SPRpcValue *	varray,
					 SPRpcValue **	v);

void		sp_rpc_value_array_set_position
					(SPRpcValue *	varray,
					 SPRpcValue *	v,
					 gint		position);

void		sp_rpc_value_array_set_position_new
					(SPRpcValue *	varray,
					 SPRpcValue **	v,
					 gint		position);

void		sp_rpc_value_array_set_before
					(SPRpcValue *	varray,
					 SPRpcValue *	v,
					 SPRpcValue *	sibling);

void		sp_rpc_value_array_set_before_new
					(SPRpcValue *	varray,
					 SPRpcValue **	v,
					 SPRpcValue *	sibling);

void		sp_rpc_value_array_remove
					(SPRpcValue *	varray,
					 SPRpcValue *	v);

void		sp_rpc_value_array_remove_by_id
					(SPRpcValue *	varray,
					 guint		id);

/* Struct: */
gboolean	sp_rpc_value_struct_has_key
					(SPRpcValue *	vstruct,
					 const gchar *	key);

SPRpcValue *	sp_rpc_value_struct_get_key
					(SPRpcValue *	vstruct,
					 const gchar *	key);

const gchar *	sp_rpc_value_struct_get_key_by_value
					(SPRpcValue *	vstruct,
					 SPRpcValue *	v);

gchar **	sp_rpc_value_struct_get_keys
					(SPRpcValue *	vstruct,
					 guint *	length)
					 G_GNUC_WARN_UNUSED_RESULT;

void		sp_rpc_value_struct_set_key
					(SPRpcValue *	vstruct,
					 const gchar *	key,
					 SPRpcValue *	v);

void		sp_rpc_value_struct_set_key_new
					(SPRpcValue *	vstruct,
					 const gchar *	key,
					 SPRpcValue **	v);

void		sp_rpc_value_struct_remove_key
					(SPRpcValue *	vstruct,
					 const gchar *	key);

void		sp_rpc_value_struct_iter_init
					(SPRpcValue *	vstruct,
					 SPRpcIter *	iter);

gboolean	sp_rpc_value_struct_iter_next
					(SPRpcIter *	iter,
					 gchar **	key,
					 SPRpcValue **	value);

void		sp_rpc_value_struct_iter_remove
					(SPRpcIter *	iter);

/* Value Builder: */
SPRpcValue *	sp_rpc_value_builder	(gchar *	format,
					 ...);

SPRpcValue *	sp_rpc_value_builder_va	(gchar *	format,
					 va_list	va);

/* Params: */
SPRpcParams *	sp_rpc_params_new	(void)
					 G_GNUC_WARN_UNUSED_RESULT;

void		sp_rpc_params_ref	(SPRpcParams *	params);
void		sp_rpc_params_unref	(SPRpcParams *	params);

void		sp_rpc_params_add	(SPRpcParams *	params,
					 const gchar *	name,
					 SPRpcValue *	value);

void		sp_rpc_params_del	(SPRpcParams *	params,
					 const gchar *	name);

gboolean	sp_rpc_params_has	(SPRpcParams *	params,
					 const gchar *	name);

SPRpcValue *	sp_rpc_params_get	(SPRpcParams *	params,
					 const gchar *	name);

gboolean	sp_rpc_params_is_empty	(SPRpcParams *	params);

void		sp_rpc_params_iter_init	(SPRpcParams *	params,
					 SPRpcIter *	iter);

gboolean	sp_rpc_params_iter_next	(SPRpcIter *	iter,
					 gchar **	name,
					 SPRpcValue **	value);

void		sp_rpc_params_iter_remove (SPRpcIter *	iter);

#endif

/* EOF */
