/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sp_rpc.h"
#include "sp_rpc_internal.h"

/**
 * sp_rpc_value_builder:
 * @format: a format string using the builder syntax
 * @...: the list of parameters to insert into the format string. 
 * @returns: a new allocated SPRpcValue, or NULL if the format is not valid
 *
 * Parses a format string and creates a valid SPRpcValue.
 */
SPRpcValue *
sp_rpc_value_builder (gchar * format, ...)
{
  SPRpcValue *value;
  va_list va;

  g_return_val_if_fail (format != NULL, NULL);

  va_start (va, format);
  value = sp_rpc_value_builder_va (format, va);
  va_end (va);

  return value;
}

static gboolean sp_rpc_value_builder_value (SPRpcValue * value,
					    gchar ** format, va_list * va);
static gboolean sp_rpc_value_builder_array (SPRpcValue * value,
					    gchar ** format, va_list * va);
static gboolean sp_rpc_value_builder_struct (SPRpcValue * value,
					     gchar ** format, va_list * va);

/**
 * sp_rpc_value_builder_va:
 * @format: a format string using the builder syntax
 * @va: the list of parameters to insert into the format string. 
 * @returns: a new allocated SPRpcValue, or NULL if the format is not valid
 *
 * Parses a format string and creates a valid SPRpcValue.
 */
SPRpcValue *
sp_rpc_value_builder_va (gchar * format, va_list va)
{
  SPRpcValue *value;

  value = sp_rpc_value_new ();

  if (sp_rpc_value_builder_value (value, &format, &va) == FALSE)
    {
      sp_rpc_value_unref (value);
      return NULL;
    }

  if (*format)
    {
      sp_rpc_value_unref (value);
      return NULL;
    }

  return value;
}

static gboolean
sp_rpc_value_builder_value (SPRpcValue * value, gchar ** format, va_list * va)
{
  switch (**format)
    {
    case 'n':
      sp_rpc_value_set_type (value, SP_RPC_TYPE_NULL);
      (*format)++;
      return TRUE;

    case 'd':
      sp_rpc_value_set_type (value, SP_RPC_TYPE_NUMBER);
      sp_rpc_value_set_number (value, va_arg (*va, gdouble));
      (*format)++;
      return TRUE;

    case 'b':
      sp_rpc_value_set_type (value, SP_RPC_TYPE_BOOLEAN);
      sp_rpc_value_set_boolean (value, va_arg (*va, gboolean));
      (*format)++;
      return TRUE;

    case 's':
      {
	gchar *tmp = va_arg (*va, gchar *);

	if (!tmp)
	  return FALSE;

	sp_rpc_value_set_type (value, SP_RPC_TYPE_STRING);
	sp_rpc_value_set_string (value, tmp);
	(*format)++;
	return TRUE;
      }

    case '(':
      return sp_rpc_value_builder_array (value, format, va);

    case '{':
      return sp_rpc_value_builder_struct (value, format, va);

    default:
      return FALSE;
    }
}

static gboolean
sp_rpc_value_builder_array (SPRpcValue * value, gchar ** format, va_list * va)
{
  sp_rpc_value_set_type (value, SP_RPC_TYPE_ARRAY);

  (*format)++;

  while (**format && **format != ')')
    {
      SPRpcValue *new;

      sp_rpc_value_array_set_new (value, &new);

      if (sp_rpc_value_builder_value (new, format, va) == FALSE)
	return FALSE;
    }

  if (**format != ')')
    return FALSE;

  (*format)++;
  return TRUE;
}

static gboolean
sp_rpc_value_builder_struct (SPRpcValue * value, gchar ** format,
			     va_list * va)
{
  sp_rpc_value_set_type (value, SP_RPC_TYPE_STRUCT);

  (*format)++;

  while (**format && **format != '}')
    {
      SPRpcValue *new;
      gchar *name;

      new = sp_rpc_value_new ();

      if (sp_rpc_value_builder_value (new, format, va) == FALSE
	  || sp_rpc_value_get_type (new) != SP_RPC_TYPE_STRING
	  || !(name = sp_rpc_value_get_string (new)))
	{
	  sp_rpc_value_unref (new);
	  return FALSE;
	}

      sp_rpc_value_unref (new);

      if (**format != ':')
	{
	  g_free (name);
	  return FALSE;
	}

      (*format)++;

      if (!**format)
	{
	  g_free (name);
	  return FALSE;
	}

      sp_rpc_value_struct_set_key_new (value, name, &new);
      g_free (name);

      if (sp_rpc_value_builder_value (new, format, va) == FALSE)
	return FALSE;

      if (**format == '}')
	break;

      if (**format != ',')
	return FALSE;

      (*format)++;

      if (!**format)
	return FALSE;
    }

  if (**format != '}')
    return FALSE;

  (*format)++;
  return TRUE;
}

/* EOF */
