/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sp_rpc.h"
#include "sp_rpc_internal.h"

#include <string.h>

/* Prototypes: */
static gboolean sp_rpc_glib_source_io_prepare (GSource * source,
					       gint * timeout_);
static gboolean sp_rpc_glib_source_io_check (GSource * source);
static gboolean sp_rpc_glib_source_io_dispatch (GSource * source,
						GSourceFunc callback,
						gpointer gdata);
static void sp_rpc_glib_source_io_finalize (GSource * source);

struct sp_rpc_glib_source_t
{
  GSource source;
  SPRpc *rpc;
};

static GSourceFuncs sp_rpc_glib_source_funcs = {
  sp_rpc_glib_source_io_prepare,
  sp_rpc_glib_source_io_check,
  sp_rpc_glib_source_io_dispatch,
  sp_rpc_glib_source_io_finalize,
  NULL,
  NULL
};

/**
 * sp_rpc_glib_source:
 * @rpc: a SPRpc
 * @returns: a GSource
 *
 * Creates a valid GSource for a GMainContext. It permits to work with a
 * GLib main loop and to receive error message trough the callback.
 */
GSource *
sp_rpc_glib_source (SPRpc * rpc)
{
  GSource *source;
  SPRpcGlibSource *rpc_source;

  g_return_val_if_fail (rpc != NULL, NULL);

  source = g_source_new (&sp_rpc_glib_source_funcs, sizeof (SPRpcGlibSource));
  rpc_source = (SPRpcGlibSource *) source;

  rpc_source->rpc = rpc;
  sp_rpc_ref (rpc);

  return (GSource *) rpc_source;
}

static gboolean
sp_rpc_glib_source_io_prepare (GSource * source, gint * timeout)
{
  gboolean pending;
  SPRpcGlibSource *s;

  s = (SPRpcGlibSource *) source;
  if (sp_rpc_event_pending (s->rpc, &pending) != SP_RPC_STATUS_DONE)
    return TRUE;

  *timeout = -1;
  return pending;
}

static gboolean
sp_rpc_glib_source_io_check (GSource * source)
{
  gboolean pending;
  SPRpcGlibSource *s;

  s = (SPRpcGlibSource *) source;
  if (sp_rpc_event_pending (s->rpc, &pending) != SP_RPC_STATUS_DONE)
    return TRUE;

  return pending;
}

static gboolean
sp_rpc_glib_source_io_dispatch (GSource * source, GSourceFunc callback,
				gpointer gdata)
{
  SPRpcGlibCallback func = (SPRpcGlibCallback) callback;
  SPRpcGlibSource *s = (SPRpcGlibSource *) source;
  SPRpcStatus status;

  if ((status = sp_rpc_iteration (s->rpc, FALSE)) == SP_RPC_STATUS_DONE)
    return TRUE;

  if (!callback)
    {
      g_warning ("SPRpc Glib Source dispatched without callback\n"
		 "You must call g_source_set_callback().");
      return FALSE;
    }

  return func (s, status, gdata);
}


static void
sp_rpc_glib_source_io_finalize (GSource * source)
{
  SPRpcGlibSource *s = (SPRpcGlibSource *) source;
  sp_rpc_unref (s->rpc);
}

/* EOF */
