/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sp_rpc.h"
#include "sp_rpc_internal.h"

#include "../json/tb_json.h"

#include <string.h>

/* WRITER ******************************************************************/
static gboolean sp_rpc_json_create_value (SPRpcValue * value,
					  tb_json_value_t * json);
static tb_json_array_t *sp_rpc_json_create_array_by_params (SPRpcParams
							    * params);
static void sp_rpc_json_create_array_by_params_cb (gchar * key,
						   SPRpcValue * value,
						   tb_json_array_t * array);

gchar *
sp_rpc_json_create (const gchar * id, guint progress_id, SPRpcJsonType type,
		    const gchar * method_name, SPRpcParams * params,
		    gsize * size)
{
  tb_json_object_t *obj_id;
  tb_json_value_t *value;
  tb_json_array_t *array;
  tb_json_object_t *obj;
  tb_json_node_t *node;
  gchar *buffer;

  if (tb_json_object_new (&obj) == FALSE)
    return NULL;

  if (tb_json_object_add_node (obj, "version", &node) == FALSE)
    goto sp_rpc_json_create_error;

  value = tb_json_node_get_value (node);
  if (tb_json_value_set_string (value, SP_RPC_PROTOCOL_VERSION) == FALSE)
    goto sp_rpc_json_create_error;

  if (tb_json_object_add_node (obj, "id", &node) == FALSE)
    goto sp_rpc_json_create_error;

  value = tb_json_node_get_value (node);
  if (tb_json_value_set_object_new (value, &obj_id) == FALSE)
    goto sp_rpc_json_create_error;

  if (tb_json_object_add_node (obj_id, "id", &node) == FALSE)
    goto sp_rpc_json_create_error;

  value = tb_json_node_get_value (node);
  if (tb_json_value_set_string (value, (gchar *) id) == FALSE)
    goto sp_rpc_json_create_error;

  if (tb_json_object_add_node (obj_id, "pid", &node) == FALSE)
    goto sp_rpc_json_create_error;

  value = tb_json_node_get_value (node);
  if (tb_json_value_set_number (value, progress_id) == FALSE)
    goto sp_rpc_json_create_error;

  if (tb_json_object_add_node (obj, "type", &node) == FALSE)
    goto sp_rpc_json_create_error;

  switch (type)
    {
    case SP_RPC_JSON_CALL:
      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, "CALL") == FALSE)
	goto sp_rpc_json_create_error;

      if (tb_json_object_add_node (obj, "method", &node) == FALSE)
	goto sp_rpc_json_create_error;

      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, (gchar *) method_name) == FALSE)
	goto sp_rpc_json_create_error;

      if (tb_json_object_add_node (obj, "params", &node) == FALSE)
	goto sp_rpc_json_create_error;

      if (!(array = sp_rpc_json_create_array_by_params (params)))
	goto sp_rpc_json_create_error;

      value = tb_json_node_get_value (node);
      if (tb_json_value_set_array (value, array) == FALSE)
	{
	  tb_json_array_destroy (array);
	  goto sp_rpc_json_create_error;
	}

      break;

    case SP_RPC_JSON_EXISTS:
      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, "EXISTS") == FALSE)
	goto sp_rpc_json_create_error;

      if (tb_json_object_add_node (obj, "method", &node) == FALSE)
	goto sp_rpc_json_create_error;

      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, (gchar *) method_name) == FALSE)
	goto sp_rpc_json_create_error;

      break;

    case SP_RPC_JSON_PARAMS:
      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, "PARAMS") == FALSE)
	goto sp_rpc_json_create_error;

      if (tb_json_object_add_node (obj, "method", &node) == FALSE)
	goto sp_rpc_json_create_error;

      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, (gchar *) method_name) == FALSE)
	goto sp_rpc_json_create_error;

      break;

    case SP_RPC_JSON_LIST:
      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, "LIST") == FALSE)
	goto sp_rpc_json_create_error;

      break;

    case SP_RPC_JSON_RESPONSE:
      value = tb_json_node_get_value (node);
      if (tb_json_value_set_string (value, "RESPONSE") == FALSE)
	goto sp_rpc_json_create_error;

      if (tb_json_object_add_node (obj, "params", &node) == FALSE)
	goto sp_rpc_json_create_error;

      if (!(array = sp_rpc_json_create_array_by_params (params)))
	goto sp_rpc_json_create_error;

      value = tb_json_node_get_value (node);
      if (tb_json_value_set_array (value, array) == FALSE)
	{
	  tb_json_array_destroy (array);
	  goto sp_rpc_json_create_error;
	}

      break;
    }

  if (tb_json_object_write_to_buffer (obj, &buffer, size, NULL) == FALSE)
    goto sp_rpc_json_create_error;

  tb_json_object_destroy (obj);
  return buffer;

sp_rpc_json_create_error:
  tb_json_object_destroy (obj);
  return NULL;
}

static tb_json_array_t *
sp_rpc_json_create_array_by_params (SPRpcParams * params)
{
  tb_json_array_t *array;

  if (tb_json_array_new (&array) == FALSE)
    return NULL;

  if (params && params->hash)
    g_hash_table_foreach (params->hash,
			  (GHFunc) sp_rpc_json_create_array_by_params_cb,
			  array);

  return array;
}

static void
sp_rpc_json_create_array_by_params_cb (gchar * key, SPRpcValue * value,
				       tb_json_array_t * array)
{
  tb_json_object_t *obj;
  tb_json_value_t *new;
  tb_json_node_t *node;
  guint id;

  if (tb_json_array_add (array, &id, &new) == FALSE)
    return;

  if (tb_json_value_set_object_new (new, &obj) == FALSE)
    goto sp_rpc_json_create_array_by_params_cb_error;

  if (tb_json_object_add_node (obj, key, &node) == FALSE)
    goto sp_rpc_json_create_array_by_params_cb_error;

  if (sp_rpc_json_create_value (value, tb_json_node_get_value (node)) ==
      FALSE)
    goto sp_rpc_json_create_array_by_params_cb_error;

  return;

sp_rpc_json_create_array_by_params_cb_error:
  tb_json_array_remove (array, id);
}

static gboolean
sp_rpc_json_create_value (SPRpcValue * value, tb_json_value_t * json)
{
  switch (sp_rpc_value_get_type (value))
    {
    case SP_RPC_TYPE_NULL:
      if (tb_json_value_set_null (json) == FALSE)
	return FALSE;

      break;

    case SP_RPC_TYPE_NUMBER:
      if (tb_json_value_set_number (json, sp_rpc_value_get_number (value)) ==
	  FALSE)
	return FALSE;
      break;

    case SP_RPC_TYPE_BOOLEAN:
      if (tb_json_value_set_boolean (json, sp_rpc_value_get_boolean (value))
	  == FALSE)
	return FALSE;
      break;

    case SP_RPC_TYPE_STRING:
      {
	gchar *tmp;
	gboolean ret;

	tmp = sp_rpc_value_get_string (value);

	if (tmp)
	  {
	    ret = tb_json_value_set_string (json, tmp);
	    g_free (tmp);
	  }
	else
	  ret = tb_json_value_set_string (json, "");

	if (ret == FALSE)
	  return FALSE;
      }
      break;

    case SP_RPC_TYPE_ARRAY:
      {
	tb_json_array_t *array;
	gint i, numb;

	if (tb_json_value_set_array_new (json, &array) == FALSE)
	  return FALSE;

	for (i = 0, numb = sp_rpc_value_array_length (value); i < numb; i++)
	  {
	    tb_json_value_t *v;

	    if (tb_json_array_add (array, NULL, &v) == FALSE)
	      return FALSE;

	    if (sp_rpc_json_create_value
		(sp_rpc_value_array_get (value, i), v) == FALSE)
	      return FALSE;
	  }

	break;
      }

    case SP_RPC_TYPE_STRUCT:
      {
	tb_json_object_t *obj;
	SPRpcIter iter;
	SPRpcValue *v;
	gchar *k;

	if (tb_json_value_set_object_new (json, &obj) == FALSE)
	  return FALSE;

	sp_rpc_value_struct_iter_init (value, &iter);
	while (sp_rpc_value_struct_iter_next (&iter, &k, &v) == TRUE)
	  {
	    tb_json_node_t *node;

	    if (tb_json_object_add_node (obj, k, &node) == FALSE)
	      return FALSE;

	    if (sp_rpc_json_create_value (v, tb_json_node_get_value (node)) ==
		FALSE)
	      return FALSE;
	  }

	break;
      }
    }

  return TRUE;
}

/* PARSER ********************************************************************/
gboolean
sp_rpc_json_parse (gchar * buffer, gsize size, gchar ** id,
		   guint * progress_id, SPRpcJsonType * type,
		   gchar ** method_name, SPRpcParams ** params)
{
  gchar *tmp;
  tb_json_t *json;
  tb_json_node_t *node;
  tb_json_object_t *obj;
  tb_json_value_t *value;
  tb_json_object_t *obj_id;

  *id = NULL;
  *method_name = NULL;
  *params = NULL;

  json = tb_json_new ();

  if (tb_json_load_from_buffer (json, buffer, size, NULL) == FALSE)
    {
      tb_json_destroy (json);
      return FALSE;
    }

  if (tb_json_is_object (json) == FALSE)
    {
      tb_json_destroy (json);
      return FALSE;
    }

  obj = tb_json_object (json);

/* X: the object, Y: string about what, Z: required? */
#define tb_json_object_get_string( x , y , z ) \
  if (!(node = tb_json_object_get_node (x, y))) \
    { \
      if (z == TRUE) \
        goto sp_rpc_json_parse_error; \
      \
      tmp = NULL; \
    } \
  else \
    { \
      value = tb_json_node_get_value (node); \
      if (tb_json_value_get_type (value) != TB_JSON_VALUE_STRING) \
        goto sp_rpc_json_parse_error; \
      \
      if (!(tmp = tb_json_value_get_string (value))) \
        goto sp_rpc_json_parse_error; \
    }

  tb_json_object_get_string (obj, "version", TRUE);

  if (g_utf8_collate (tmp, SP_RPC_PROTOCOL_VERSION))
    {
      g_free (tmp);
      goto sp_rpc_json_parse_error;
    }

  g_free (tmp);

  if (!(node = tb_json_object_get_node (obj, "id")))
    goto sp_rpc_json_parse_error;

  value = tb_json_node_get_value (node);
  if (tb_json_value_get_type (value) != TB_JSON_VALUE_OBJECT)
    goto sp_rpc_json_parse_error;

  obj_id = tb_json_value_get_object (value);
  tb_json_object_get_string (obj_id, "id", TRUE);
  *id = tmp;

  if (!(node = tb_json_object_get_node (obj_id, "pid")))
    goto sp_rpc_json_parse_error;

  value = tb_json_node_get_value (node);
  if (tb_json_value_get_type (value) != TB_JSON_VALUE_NUMBER)
    goto sp_rpc_json_parse_error;

  *progress_id = tb_json_value_get_number (value);

  tb_json_object_get_string (obj, "type", TRUE);

  if (!strcmp (tmp, "CALL"))
    *type = SP_RPC_JSON_CALL;

  else if (!strcmp (tmp, "EXISTS"))
    *type = SP_RPC_JSON_EXISTS;

  else if (!strcmp (tmp, "PARAMS"))
    *type = SP_RPC_JSON_PARAMS;

  else if (!strcmp (tmp, "LIST"))
    *type = SP_RPC_JSON_LIST;

  else if (!strcmp (tmp, "RESPONSE"))
    *type = SP_RPC_JSON_RESPONSE;

  else
    {
      g_free (tmp);
      goto sp_rpc_json_parse_error;
    }

  g_free (tmp);

  tb_json_object_get_string (obj, "method", FALSE);
  *method_name = tmp;

  if ((node = tb_json_object_get_node (obj, "params"))
      &&
      sp_rpc_params_new_from_json_value (tb_json_node_get_value (node),
					 params) == FALSE)
    goto sp_rpc_json_parse_error;

  tb_json_destroy (json);
  return TRUE;

sp_rpc_json_parse_error:
  tb_json_destroy (json);

  if (*id)
    g_free (*id);

  if (*method_name)
    g_free (*method_name);

  if (*params)
    sp_rpc_params_unref (*params);

  return FALSE;
}

/* EOF */
