/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sp_rpc.h"
#include "sp_rpc_internal.h"

/**
 * sp_rpc_params_new:
 * @returns: a new SPRpcParams with a reference count of 1
 *
 * Creates a new SPRpcParams.
 */
SPRpcParams *
sp_rpc_params_new (void)
{
  SPRpcParams *ret;

  ret = g_malloc0 (sizeof (SPRpcParams));

  ret->hash =
    g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
			   (GDestroyNotify) sp_rpc_value_unref);

  sp_rpc_params_ref (ret);
  return ret;
}

static SPRpcValue *sp_rpc_params_new_from_json_value_real (tb_json_value_t *
							   value);

gboolean
sp_rpc_params_new_from_json_value (tb_json_value_t * value,
				   SPRpcParams ** params)
{
  guint i, numb;
  SPRpcParams *mp;
  tb_json_array_t *array;

  if (tb_json_value_get_type (value) != TB_JSON_VALUE_ARRAY)
    return FALSE;

  array = tb_json_value_get_array (value);
  numb = tb_json_array_length (array);

  mp = sp_rpc_params_new ();

  for (i = 0; i < numb; i++)
    {
      tb_json_value_t *value;
      tb_json_object_t *obj;
      SPRpcValue *v;
      GList *nodes;
      gchar *tmp;

      value = tb_json_array_get (array, i);
      if (tb_json_value_get_type (value) != TB_JSON_VALUE_OBJECT)
	{
	  sp_rpc_params_unref (mp);
	  return FALSE;
	}

      obj = tb_json_value_get_object (value);
      nodes = tb_json_object_get_nodes (obj);

      if (!nodes || nodes->next)
	{
	  sp_rpc_params_unref (mp);
	  return FALSE;
	}

      if (!
	  (v =
	   sp_rpc_params_new_from_json_value_real
	   (tb_json_node_get_value (nodes->data))))
	{
	  sp_rpc_params_unref (mp);
	  return FALSE;
	}

      tmp = tb_json_node_get_string (nodes->data);
      sp_rpc_params_add (mp, tmp, v);
      sp_rpc_value_unref (v);
      g_free (tmp);
    }

  *params = mp;
  return TRUE;
}

static SPRpcValue *
sp_rpc_params_new_from_json_value_real (tb_json_value_t * value)
{
  SPRpcValue *v;

  v = sp_rpc_value_new ();

  switch (tb_json_value_get_type (value))
    {
    case TB_JSON_VALUE_NULL:
      sp_rpc_value_set_type (v, SP_RPC_TYPE_NULL);
      break;

    case TB_JSON_VALUE_STRING:
      {
	gchar *tmp;

	sp_rpc_value_set_type (v, SP_RPC_TYPE_STRING);
	tmp = tb_json_value_get_string (value);
	sp_rpc_value_set_string (v, tmp);
	g_free (tmp);
      }
      break;

    case TB_JSON_VALUE_NUMBER:
      sp_rpc_value_set_type (v, SP_RPC_TYPE_NUMBER);
      sp_rpc_value_set_number (v, tb_json_value_get_number (value));
      break;

    case TB_JSON_VALUE_OBJECT:
      {
	GList *nodes;
	tb_json_object_t *object;

	sp_rpc_value_set_type (v, SP_RPC_TYPE_STRUCT);
	object = tb_json_value_get_object (value);

	for (nodes = tb_json_object_get_nodes (object); nodes;
	     nodes = nodes->next)
	  {
	    tb_json_node_t *node = nodes->data;
	    SPRpcValue *new;
	    gchar *tmp;

	    new =
	      sp_rpc_params_new_from_json_value_real
	      (tb_json_node_get_value (node));

	    tmp = tb_json_node_get_string (node);
	    sp_rpc_value_struct_set_key (v, tmp, new);
	    g_free (tmp);

	    sp_rpc_value_unref (new);
	  }
      }
      break;

    case TB_JSON_VALUE_ARRAY:
      {
	guint i, numb;
	tb_json_array_t *array;

	sp_rpc_value_set_type (v, SP_RPC_TYPE_ARRAY);

	array = tb_json_value_get_array (value);
	numb = tb_json_array_length (array);

	for (i = 0; i < numb; i++)
	  {
	    SPRpcValue *new;

	    new =
	      sp_rpc_params_new_from_json_value_real (tb_json_array_get
						      (array, i));

	    sp_rpc_value_array_set (v, new);
	    sp_rpc_value_unref (new);
	  }
      }
      break;

    case TB_JSON_VALUE_BOOLEAN:
      sp_rpc_value_set_type (v, SP_RPC_TYPE_BOOLEAN);
      sp_rpc_value_set_boolean (v, tb_json_value_get_boolean (value));
      break;
    }

  return v;
}

SPRpcParams *
sp_rpc_params_new_for_call_list (GList * methods)
{
  SPRpcParams *params;

  params = sp_rpc_params_new ();

  for (; methods; methods = methods->next)
    {
      SPRpcCb *method = methods->data;
      SPRpcValue *value, *v;

      value = sp_rpc_value_new ();
      sp_rpc_value_set_type (value, SP_RPC_TYPE_STRUCT);
      sp_rpc_params_add (params, method->name, value);
      sp_rpc_value_unref (value);

      if (method->description)
	{
	  v = sp_rpc_value_new ();
	  sp_rpc_value_set_type (v, SP_RPC_TYPE_STRING);
	  sp_rpc_value_set_string (v, method->description);

	  sp_rpc_value_struct_set_key (value, "description", v);
	  sp_rpc_value_unref (v);
	}

      if (method->params)
	{
	  GHashTableIter iter;
	  gpointer pk, pv;

	  v = sp_rpc_value_new ();
	  sp_rpc_value_set_type (v, SP_RPC_TYPE_ARRAY);

	  g_hash_table_iter_init (&iter, method->params->hash);
	  while (g_hash_table_iter_next (&iter, &pk, &pv))
	    {
	      SPRpcValue *n;

	      n = sp_rpc_value_new ();
	      sp_rpc_value_set_type (n, SP_RPC_TYPE_STRUCT);
	      sp_rpc_value_struct_set_key (n, pk, pv);

	      sp_rpc_value_array_set (v, n);
	      sp_rpc_value_unref (n);
	    }

	  sp_rpc_value_struct_set_key (value, "params", v);
	  sp_rpc_value_unref (v);
	}
    }

  return params;
}

SPRpcParams *
sp_rpc_params_new_for_call_exists (gchar * method)
{
  SPRpcParams *params;
  SPRpcValue *value;

  params = sp_rpc_params_new ();

  value = sp_rpc_value_new ();
  sp_rpc_value_set_type (value, SP_RPC_TYPE_BOOLEAN);
  sp_rpc_value_set_boolean (value, TRUE);
  sp_rpc_params_add (params, "exists", value);

  return params;
}

/**
 * sp_rpc_params_ref:
 * @params: a SPRpcParams
 *
 * Increases the reference count of a SPRpcParams.
 */
void
sp_rpc_params_ref (SPRpcParams * params)
{
  g_return_if_fail (params != NULL);

  params->ref++;
}

/**
 * sp_rpc_params_unref:
 * @params: a SPRpcParams
 *
 * Decreases the reference count of a SPRpcParams. When its reference count
 * drops to 0, the SPRpcParams is finalized.
 */
void
sp_rpc_params_unref (SPRpcParams * params)
{
  g_return_if_fail (params != NULL);

  params->ref--;

  if (params->ref > 0)
    return;

  if (params->hash)
    g_hash_table_destroy (params->hash);

  g_free (params);
}

/**
 * sp_rpc_params_add:
 * @params: a SPRpcParams
 * @name: the name of the new value
 * @value: a SPRpcValue
 *
 * Inserts into the params array a new value and increases its count 
 * references.
 */
void
sp_rpc_params_add (SPRpcParams * params, const gchar * name,
		   SPRpcValue * value)
{
  g_return_if_fail (params != NULL);
  g_return_if_fail (name != NULL);
  g_return_if_fail (value != NULL);
  g_return_if_fail (sp_rpc_params_has (params, name) == FALSE);

  g_hash_table_insert (params->hash, g_strdup (name), value);
  sp_rpc_value_ref (value);
}

/**
 * sp_rpc_params_del:
 * @params: a SPRpcParams
 * @name: the name of the value
 *
 * Removes a value from a SPRpcParams and decreases the
 * count reference of this value
 */
void
sp_rpc_params_del (SPRpcParams * params, const gchar * name)
{
  g_return_if_fail (params != NULL);
  g_return_if_fail (name != NULL);
  g_return_if_fail (sp_rpc_params_has (params, name) == TRUE);

  g_hash_table_remove (params->hash, name);
}

/**
 * sp_rpc_params_has:
 * @params: a SPRpcParams
 * @name: the name of the value
 * @returns: TRUE if key is a part of the params, FALSE otherwise. 
 *
 * Looks whether the params array has the value @name.
 */
gboolean
sp_rpc_params_has (SPRpcParams * params, const gchar * name)
{
  g_return_val_if_fail (params != NULL, FALSE);
  g_return_val_if_fail (name != NULL, FALSE);

  if (g_hash_table_lookup (params->hash, name))
    return TRUE;

  return FALSE;
}

/**
 * sp_rpc_params_get:
 * @params: a SPRpcParams
 * @name: the name of the value
 * @returns: the associated value, or NULL
 *
 * Looks up a value in a params array.
 */
SPRpcValue *
sp_rpc_params_get (SPRpcParams * params, const gchar * name)
{
  g_return_val_if_fail (params != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (sp_rpc_params_has (params, name) == TRUE, NULL);

  return g_hash_table_lookup (params->hash, name);
}

/**
 * sp_rpc_params_is_empty:
 * @params: a SPRpcParams
 * @return: TRUE if the params array is empty
 *
 * Returns TRUE if the params array is empty.
 */
gboolean
sp_rpc_params_is_empty (SPRpcParams * params)
{
  GHashTableIter iter;
  gpointer k, v;

  g_return_val_if_fail (params != NULL, FALSE);

  g_hash_table_iter_init (&iter, params->hash);
  return !g_hash_table_iter_next (&iter, &k, &v);
}

static gboolean sp_rpc_params_validate_value (SPRpcValue * mvalue,
					      SPRpcValue * pvalue);

SPRpcParams *
sp_rpc_params_validate (SPRpcParams * model, SPRpcParams * params)
{
  SPRpcParams *new;
  GHashTableIter iter;
  gpointer key, value;

  SPRpcValue *params_value;
  SPRpcValue *model_value;

  if (!model)
    return NULL;

  new = sp_rpc_params_new ();

  g_hash_table_iter_init (&iter, model->hash);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      model_value = value;
      params_value = g_hash_table_lookup (params->hash, key);

      if (!params_value
	  || sp_rpc_params_validate_value (model_value,
					   params_value) == FALSE)
	sp_rpc_params_add (new, key, model_value);

      else
	sp_rpc_params_add (new, key, params_value);
    }

  return new;
}

static gboolean
sp_rpc_params_validate_value (SPRpcValue * mvalue, SPRpcValue * pvalue)
{
  switch (sp_rpc_value_get_type (mvalue))
    {
    case SP_RPC_TYPE_NULL:
      return TRUE;

    case SP_RPC_TYPE_NUMBER:
    case SP_RPC_TYPE_BOOLEAN:
    case SP_RPC_TYPE_STRING:
      if (sp_rpc_value_get_type (pvalue) == sp_rpc_value_get_type (mvalue))
	return TRUE;
      return FALSE;

    case SP_RPC_TYPE_ARRAY:
      {
	guint pid, mnumb, pnumb;

	if (sp_rpc_value_get_type (pvalue) != SP_RPC_TYPE_ARRAY)
	  return FALSE;

	if (!(mnumb = sp_rpc_value_array_length (mvalue)))
	  return TRUE;

	if (!(pnumb = sp_rpc_value_array_length (pvalue)))
	  return FALSE;

	for (pid = 0; (pnumb - pid) >= mnumb; pid += mnumb)
	  {
	    gint i;

	    for (i = 0; i < mnumb; i++)
	      {
		SPRpcValue *mv, *pv;

		mv = sp_rpc_value_array_get (mvalue, i);
		pv = sp_rpc_value_array_get (pvalue, pid + i);

		if (sp_rpc_params_validate_value (mv, pv) == FALSE)
		  return FALSE;
	      }
	  }

	if (pid == pnumb)
	  return TRUE;

	return FALSE;
      }

    case SP_RPC_TYPE_STRUCT:
      {
	gchar **mkeys;
	gchar **pkeys;
	gsize ml, pl;

	if (!(mkeys = sp_rpc_value_struct_get_keys (mvalue, &ml)))
	  return FALSE;

	if (!(pkeys = sp_rpc_value_struct_get_keys (pvalue, &pl)))
	  {
	    g_strfreev (mkeys);
	    return FALSE;
	  }

	if (ml != pl)
	  {
	    g_strfreev (mkeys);
	    g_strfreev (pkeys);
	    return FALSE;
	  }

	for (ml = 0; mkeys[ml]; ml++)
	  {
	    SPRpcValue *mv, *pv;

	    mv = sp_rpc_value_struct_get_key (mvalue, mkeys[ml]);

	    if (!(pv = sp_rpc_value_struct_get_key (pvalue, mkeys[ml]))
		|| sp_rpc_params_validate_value (mv, pv) == FALSE)
	      {
		g_strfreev (mkeys);
		g_strfreev (pkeys);
		return FALSE;
	      }
	  }

	g_strfreev (mkeys);
	g_strfreev (pkeys);
	return TRUE;
      }
    }

  return FALSE;
}

/**
 * sp_rpc_params_iter_init:
 * @params: a SPRpcParams
 * @iter: an unitialized SPRpcIter
 *
 * Initializes a key/value pair iterator and associates it with the a
 * SPRpcParams.
 * Modifying the struct after calling this function invalidates the returned
 * iterator.
 *
 * |[
 * gchar *key;
 * SPRpcValue *value;
 * SPRpcIter iter;
 *
 * sp_rpc_params_iter_initt(params, &iter);
 * while (sp_rpc_params_iter_next(&iter, &key, &value)) 
 *   {
 *     /&ast; do something with key and value &ast;/
 *   }
 * ]|
 */
void
sp_rpc_params_iter_init (SPRpcParams * params, SPRpcIter * iter)
{
  g_return_if_fail (params != NULL);
  g_return_if_fail (iter != NULL);

  g_hash_table_iter_init (iter, params->hash);
}

/**
 * sp_rpc_params_iter_next:
 * @iter: an initialized SPRpcIter
 * @name: a location for the key, or NULL
 * @value: the location for the value, or NULL
 * @returns: FALSE if the end of struct has been reached
 *
 * Advances iter and retrieves the key and/or value that are now pointed to
 * as a result of this advancement. If FALSE is returned, key and value are
 * not set, and the iterator becomes invalid.
 */
gboolean
sp_rpc_params_iter_next (SPRpcIter * iter, gchar ** name, SPRpcValue ** value)
{
  g_return_val_if_fail (iter != NULL, FALSE);

  return g_hash_table_iter_next (iter, (gpointer *) name, (gpointer *) value);
}

/**
 * sp_rpc_params_iter_remove:
 * @iter: an initialized SPRpcIter
 *
 * Removes the key/value pair currently pointed to by the iterator from its
 * associated SPRpcValue. Can only be called after
 * sp_rpc_params_iter_next() returned TRUE, and cannot be called more
 * than once for the same key/value pair. 
 *
 * The count reference of the removed value will be decreased.
 */
void
sp_rpc_params_iter_remove (SPRpcIter * iter)
{
  g_return_if_fail (iter != NULL);

  g_hash_table_iter_remove (iter);
}

/* EOF */
