/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sp_rpc.h"
#include "sp_rpc_internal.h"

/**
 * sp_rpc_value_new:
 * @returns: a new SPRpcValue with a reference count of 1
 *
 * Creates a new SPRpcValue with type SP_RPC_TYPE_NULL.
 */
SPRpcValue *
sp_rpc_value_new (void)
{
  SPRpcValue *value;

  value = g_malloc0 (sizeof (SPRpcValue));
  value->type = SP_RPC_TYPE_NULL;

  sp_rpc_value_ref (value);
  return value;
}

/**
 * sp_rpc_value_ref:
 * @value: a SPRpcValue
 *
 * Increases the reference count of a SPRpcValue.
 */
void
sp_rpc_value_ref (SPRpcValue * value)
{
  g_return_if_fail (value != NULL);

  value->ref++;
}

static void
sp_rpc_value_reset (SPRpcValue * value)
{
  if (!value)
    return;

  switch (value->type)
    {
    case SP_RPC_TYPE_NULL:
      break;

    case SP_RPC_TYPE_NUMBER:
      value->v_number = 0;
      break;

    case SP_RPC_TYPE_BOOLEAN:
      value->v_boolean = 0;
      break;

    case SP_RPC_TYPE_STRING:
      if (value->v_string)
	{
	  g_free (value->v_string);
	  value->v_string = NULL;
	}
      break;

    case SP_RPC_TYPE_ARRAY:
      if (value->v_array.array)
	{
	  g_list_foreach (value->v_array.array, (GFunc) sp_rpc_value_unref,
			  NULL);
	  g_list_free (value->v_array.array);
	  value->v_array.array = NULL;
	  value->v_array.array_length = 0;
	}
      break;

    case SP_RPC_TYPE_STRUCT:
      if (value->v_struct)
	{
	  g_hash_table_destroy (value->v_struct);
	  value->v_struct = NULL;
	}
    }
}

/**
 * sp_rpc_value_unref:
 * @value: a SPRpcValue
 *
 * Decreases the reference count of a SPRpcValue. When its reference count
 * drops to 0, the SPRpcValue is finalized.
 */
void
sp_rpc_value_unref (SPRpcValue * value)
{
  g_return_if_fail (value != NULL);

  value->ref--;

  if (value->ref > 0)
    return;

  sp_rpc_value_reset (value);
  g_free (value);
}

/**
 * sp_rpc_value_set_type:
 * @value: a SPRpcValue
 * @type: the type
 *
 * Changes the type of a SPRpcValue. It removes the previous value of the
 * SPRpcValue.
 */
void
sp_rpc_value_set_type (SPRpcValue * value, SPRpcTypeValue type)
{
  g_return_if_fail (value != NULL);

  sp_rpc_value_reset (value);
  value->type = type;

  switch (type)
    {
    case SP_RPC_TYPE_NULL:
    case SP_RPC_TYPE_NUMBER:
    case SP_RPC_TYPE_BOOLEAN:
    case SP_RPC_TYPE_STRING:
    case SP_RPC_TYPE_ARRAY:
      break;

    case SP_RPC_TYPE_STRUCT:
      value->v_struct =
	g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
			       (GDestroyNotify) sp_rpc_value_unref);
      break;
    }
}

/**
 * sp_rpc_value_get_type:
 * @value: a SPRpcValue
 * @returns: the type of the SPRpcValue
 *
 * Returns the type of a SPRpcValue
 */
SPRpcTypeValue
sp_rpc_value_get_type (SPRpcValue * value)
{
  g_return_val_if_fail (value != NULL, 0);
  return value->type;
}

/* Simple Values ***********************************************************/

/**
 * sp_rpc_value_set_number:
 * @value: a SPRpcValue
 * @v_number: the number of the SPRpcValue
 *
 * Sets the @v_number into a SPRpcValue if its type is SP_RPC_TYPE_NUMBER.
 */
void
sp_rpc_value_set_number (SPRpcValue * value, gdouble v_number)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_NUMBER);

  value->v_number = v_number;
}

/**
 * sp_rpc_value_get_number:
 * @value: a SPRpcValue
 * @returns: the value of the SPRpcValue
 *
 * Returns the value of SPRpcValue if its type is SP_RPC_TYPE_NUMBER.
 */
gdouble
sp_rpc_value_get_number (SPRpcValue * value)
{
  g_return_val_if_fail (value != NULL, 0);
  g_return_val_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_NUMBER,
			0);

  return value->v_number;
}

/**
 * sp_rpc_value_set_boolean:
 * @value: a SPRpcValue
 * @v_boolean: a boolean value
 *
 * Sets a boolean value into a SPRpcValue if its type is SP_RPC_TYPE_BOOLEAN.
 */
void
sp_rpc_value_set_boolean (SPRpcValue * value, gboolean v_boolean)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_BOOLEAN);

  value->v_boolean = v_boolean;
}

/**
 * sp_rpc_value_get_boolean:
 * @value: a SPRpcValue
 * @returns: the value of the SPRpcValue
 *
 * Returns the value of SPRpcValue if its type is SP_RPC_TYPE_BOOLEAN.
 */
gboolean
sp_rpc_value_get_boolean (SPRpcValue * value)
{
  g_return_val_if_fail (value != NULL, FALSE);
  g_return_val_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_BOOLEAN,
			FALSE);

  return value->v_boolean;
}

/**
 * sp_rpc_value_set_string:
 * @value: a SPRpcValue
 * @v_string: a string
 *
 * Sets a string into a SPRpcValue if its type is SP_RPC_TYPE_STRING.
 */
void
sp_rpc_value_set_string (SPRpcValue * value, gchar * v_string)
{
  g_return_if_fail (value != NULL);
  g_return_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_STRING);

  if (value->v_string)
    g_free (value->v_string);

  if (v_string)
    value->v_string = g_strdup (v_string);
  else
    value->v_string = NULL;
}

/**
 * sp_rpc_value_set_string_printf:
 * @value: a SPRpcValue
 * @format: a format string
 * @...: arguments of the format string
 *
 * Sets a string from a format string into a SPRpcValue if its type
 * is SP_RPC_TYPE_STRING.
 */
void
sp_rpc_value_set_string_printf (SPRpcValue * value, gchar * format, ...)
{
  va_list va;
  gchar *string;

  g_return_if_fail (value != NULL);
  g_return_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_STRING);

  va_start (va, format);
  string = g_strdup_vprintf (format, va);
  va_end (va);

  sp_rpc_value_set_string (value, string);
  g_free (string);
}

/**
 * sp_rpc_value_get_string:
 * @value: a SPRpcValue
 * @returns: a new allocated string
 *
 * Returns the value of SPRpcValue if its type is SP_RPC_TYPE_STRING.
 * This string must be freed with g_free().
 */
gchar *
sp_rpc_value_get_string (SPRpcValue * value)
{
  g_return_val_if_fail (value != NULL, NULL);
  g_return_val_if_fail (sp_rpc_value_get_type (value) == SP_RPC_TYPE_STRING,
			NULL);

  if (value->v_string)
    return g_strdup (value->v_string);

  return NULL;
}

/* Arrays *******************************************************************/

/**
 * sp_rpc_value_array_length:
 * @varray: a SPRpcValue
 * @returns: the number of items of an array
 *
 * Returns the number of items of an array if the @varray is a
 * SP_RPC_TYPE_ARRAY.
 */
guint
sp_rpc_value_array_length (SPRpcValue * varray)
{
  g_return_val_if_fail (varray != NULL, 0);
  g_return_val_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY,
			0);

  return varray->v_array.array_length;
}

/**
 * sp_rpc_value_array_get:
 * @varray: a SPRpcValue
 * @id: the position of the item
 * @returns: a SPRpcValue in the position id
 *
 * Returns an item of an array if the @varray is a SP_RPC_TYPE_ARRAY.
 */
SPRpcValue *
sp_rpc_value_array_get (SPRpcValue * varray, guint id)
{
  g_return_val_if_fail (varray != NULL, NULL);
  g_return_val_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY,
			NULL);
  g_return_val_if_fail (sp_rpc_value_array_length (varray) >= id, NULL);


  return g_list_nth_data (varray->v_array.array, id);
}

/**
 * sp_rpc_value_array_get_id_by_value
 * @varray: a SPRpcValue
 * @v: a SPRpcValue
 * @returns: the position of the value v into the array, or -1
 *
 * Returns the position of the value @v into the @varray or -1 if it is not
 * into the array.
 */
gint
sp_rpc_value_array_get_id_by_value (SPRpcValue * varray, SPRpcValue * v)
{
  g_return_val_if_fail (varray != NULL, -1);
  g_return_val_if_fail (v != NULL, -1);
  g_return_val_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY,
			-1);

  return g_list_index (varray->v_array.array, v);
}

/**
 * sp_rpc_value_array_set:
 * @varray: a SPRpcValue
 * @v: a SPRpcValue
 *
 * Inserts a new value into the @varray. The count references of the value
 * will be increased.
 */
void
sp_rpc_value_array_set (SPRpcValue * varray, SPRpcValue * v)
{
  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  sp_rpc_value_array_set_position (varray, v, -1);
}

/**
 * sp_rpc_value_array_set_new:
 * @varray: a SPRpcValue
 * @v: the location of a SPRpcValue
 *
 * Creates and inserts a value into an varray.
 */
void
sp_rpc_value_array_set_new (SPRpcValue * varray, SPRpcValue ** v)
{
  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  *v = sp_rpc_value_new ();
  sp_rpc_value_array_set_position (varray, *v, -1);

  sp_rpc_value_unref (*v);
}

/**
 * sp_rpc_value_array_set_position:
 * @varray: a SPRpcValue
 * @v: a SPRpcValue
 * @position: the position to insert the element. If this is negative, or
 * is larger than the number of elements in the array, the new element is
 * added on to the end of the array. 
 *
 * Inserts a new value into the @varray at the given position.
 */
void
sp_rpc_value_array_set_position (SPRpcValue * varray, SPRpcValue * v,
				 gint position)
{
  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  varray->v_array.array = g_list_insert (varray->v_array.array, v, position);
  varray->v_array.array_length++;

  sp_rpc_value_ref (v);
}

/**
 * sp_rpc_value_array_set_position_new:
 * @varray: a SPRpcValue
 * @v: the location of a SPRpcValue
 * @position: the position to insert the element. If this is negative, or
 * is larger than the number of elements in the array, the new element is
 * added on to the end of the array. 
 *
 * Creates and inserts a new value into the array at the given position.
 */
void
sp_rpc_value_array_set_position_new (SPRpcValue * varray, SPRpcValue ** v,
				     gint position)
{
  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  *v = sp_rpc_value_new ();
  sp_rpc_value_array_set_position (varray, *v, position);

  sp_rpc_value_unref (*v);
}

/**
 * sp_rpc_value_array_set_before:
 * @varray: a SPRpcValue
 * @v: a SPRpcValue
 * @sibling: the array item before which the new value is inserted or NULL
 * to insert at the end of the array;
 *
 * Inserts a new value into the array before the given position.
 */
void
sp_rpc_value_array_set_before (SPRpcValue * varray, SPRpcValue * v,
			       SPRpcValue * sibling)
{
  GList *list;

  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  if (!sibling)
    list = NULL;

  else
    list = g_list_find (varray->v_array.array, sibling);

  varray->v_array.array =
    g_list_insert_before (varray->v_array.array, list, v);
  varray->v_array.array_length++;

  sp_rpc_value_ref (v);
}

/**
 * sp_rpc_value_array_set_before_new:
 * @varray: a SPRpcValue
 * @v: the location of a SPRpcValue
 * @sibling: the array item before which the new value is inserted or NULL
 * to insert at the end of the array;
 *
 * Creates and inserts a new value into the array before the given position.
 */
void
sp_rpc_value_array_set_before_new (SPRpcValue * varray, SPRpcValue ** v,
				   SPRpcValue * sibling)
{
  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  *v = sp_rpc_value_new ();
  sp_rpc_value_array_set_before (varray, *v, sibling);

  sp_rpc_value_unref (*v);
}

/**
 * sp_rpc_value_array_remove:
 * @varray: a SPRpcValue
 * @v: a SPRpcValue
 *
 * Removes a value from the array and decreases its reference count.
 */
void
sp_rpc_value_array_remove (SPRpcValue * varray, SPRpcValue * v)
{
  g_return_if_fail (varray != NULL);
  g_return_if_fail (v != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);

  sp_rpc_value_unref (v);
  varray->v_array.array = g_list_remove (varray->v_array.array, v);
  varray->v_array.array_length--;
}

/**
 * sp_rpc_value_array_remove_by_id:
 * @varray: a SPRpcValue
 * @id: the position
 *
 * Removes a value by its position from the array and decreases its
 * reference count.
 */
void
sp_rpc_value_array_remove_by_id (SPRpcValue * varray, guint id)
{
  SPRpcValue *v;

  g_return_if_fail (varray != NULL);
  g_return_if_fail (sp_rpc_value_get_type (varray) == SP_RPC_TYPE_ARRAY);
  g_return_if_fail (sp_rpc_value_array_length (varray) >= id);

  if (!(v = g_list_nth_data (varray->v_array.array, id)))
    return;

  sp_rpc_value_array_remove (varray, v);
}

/* Struct ********************************************************************/

/**
 * sp_rpc_value_struct_has_key:
 * @vstruct: a SPRpcValue
 * @key: the key
 * @returns: TRUE if key is a part of the struct, FALSE otherwise. 
 *
 * Looks whether the struct has the key @key.
 */
gboolean
sp_rpc_value_struct_has_key (SPRpcValue * vstruct, const gchar * key)
{
  g_return_val_if_fail (vstruct != NULL, FALSE);
  g_return_val_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT,
			FALSE);
  g_return_val_if_fail (key != NULL, FALSE);

  if (g_hash_table_lookup (vstruct->v_struct, key))
    return TRUE;

  return FALSE;
}

/**
 * sp_rpc_value_struct_get_key:
 * @vstruct: a SPRpcValue
 * @key: the key
 * @returns: the value of a key
 *
 * returns the value of a key
 */
SPRpcValue *
sp_rpc_value_struct_get_key (SPRpcValue * vstruct, const gchar * key)
{
  SPRpcValue *v;

  g_return_val_if_fail (vstruct != NULL, NULL);
  g_return_val_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT,
			NULL);
  g_return_val_if_fail (key != NULL, NULL);

  if (!(v = g_hash_table_lookup (vstruct->v_struct, key)))
    g_return_val_if_fail (sp_rpc_value_struct_has_key (vstruct, key) == TRUE,
			  NULL);

  return v;
}

/**
 * sp_rpc_value_struct_get_key_by_value:
 * @vstruct: a SPRpcValue
 * @v: a SPRpcValue
 * @returns: the key of a value
 *
 * Returns the key of a value, or NULL if the value @v is not into the
 * struct. The string is internal and you should not free it.
 */
const gchar *
sp_rpc_value_struct_get_key_by_value (SPRpcValue * vstruct, SPRpcValue * v)
{
  GHashTableIter iter;
  gpointer pk, pv;

  g_return_val_if_fail (vstruct != NULL, NULL);
  g_return_val_if_fail (v != NULL, NULL);
  g_return_val_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT,
			NULL);

  g_hash_table_iter_init (&iter, vstruct->v_struct);
  while (g_hash_table_iter_next (&iter, &pk, &pv))
    if (pv == v)
      return (const gchar *) pk;

  return NULL;
}

/**
 * sp_rpc_value_struct_get_keys:
 * @vstruct: a SPRpcValue
 * @length: the location of the length of the array of keys, or NULL
 * @returns: a new allocated array of the keys
 *
 * creates an array of the keys of a struct. Use g_strfreev() after the usage.
 */
gchar **
sp_rpc_value_struct_get_keys (SPRpcValue * vstruct, guint * length)
{
  GList *keys;
  GList *list;
  gchar **ret;
  gint id;

  g_return_val_if_fail (vstruct != NULL, NULL);
  g_return_val_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT,
			NULL);

  if (!(keys = g_hash_table_get_keys (vstruct->v_struct)))
    return NULL;

  ret = g_malloc (sizeof (gchar *) * (g_list_length (keys) + 1));
  for (list = keys, id = 0; list; list = list->next, id++)
    ret[id] = g_strdup (list->data);

  g_list_free (keys);
  ret[id] = NULL;

  if (length)
    *length = id;

  return ret;
}

/**
 * sp_rpc_value_struct_set_key:
 * @vstruct: a SPRpcValue
 * @key: the key of the new struct
 * @v: a SPRpcValue
 *
 * Inserts a new key and value into a struct.
 *
 * If the key already exists into the struct, its current value is replaced
 * with the new value and the old value will be unreferenced.
 *
 * The count reference of the @v will be increased.
 */
void
sp_rpc_value_struct_set_key (SPRpcValue * vstruct, const gchar * key,
			     SPRpcValue * v)
{
  g_return_if_fail (vstruct != NULL);
  g_return_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT);
  g_return_if_fail (key != NULL);
  g_return_if_fail (v != NULL);

  g_hash_table_insert (vstruct->v_struct, g_strdup (key), v);
  sp_rpc_value_ref (v);
}

/**
 * sp_rpc_value_struct_set_key_new:
 * @vstruct: a SPRpcValue
 * @key: the key of the new struct
 * @v: the location of the new SPRpcValue
 *
 * Creates and inserts a new key and value into a struct.
 * Read how the sp_rpc_value_struct_set_key() works.
 */
void
sp_rpc_value_struct_set_key_new (SPRpcValue * vstruct, const gchar * key,
				 SPRpcValue ** v)
{
  g_return_if_fail (vstruct != NULL);
  g_return_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT);
  g_return_if_fail (key != NULL);
  g_return_if_fail (v != NULL);

  *v = sp_rpc_value_new ();
  sp_rpc_value_struct_set_key (vstruct, key, *v);
  sp_rpc_value_unref (*v);
}

/**
 * sp_rpc_value_struct_remove_key:
 * @vstruct: a SPRpcValue
 * @key: the key
 *
 * Removes a key and its associated value from a struct.
 * The count reference of the value will be decreased.
 */
void
sp_rpc_value_struct_remove_key (SPRpcValue * vstruct, const gchar * key)
{
  g_return_if_fail (vstruct != NULL);
  g_return_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT);
  g_return_if_fail (key != NULL);

  g_hash_table_remove (vstruct->v_struct, key);
}

/**
 * sp_rpc_value_struct_iter_init:
 * @vstruct: a SPRpcValue
 * @iter: an unitialized SPRpcIter
 *
 * Initializes a key/value pair iterator and associates it with the vstruct.
 * Modifying the struct after calling this function invalidates the returned
 * iterator.
 *
 * |[
 * gchar *key;
 * SPRpcValue *value;
 * SPRpcIter iter;
 *
 * sp_rpc_value_struct_iter_init(vstruct, &iter);
 * while (sp_rpc_value_struct_iter_next(&iter, &key, &value)) 
 *   {
 *     /&ast; do something with key and value &ast;/
 *   }
 * ]|
 */
void
sp_rpc_value_struct_iter_init (SPRpcValue * vstruct, SPRpcIter * iter)
{
  g_return_if_fail (vstruct != NULL);
  g_return_if_fail (sp_rpc_value_get_type (vstruct) == SP_RPC_TYPE_STRUCT);
  g_return_if_fail (iter != NULL);

  g_hash_table_iter_init (iter, vstruct->v_struct);
}

/**
 * sp_rpc_value_struct_iter_next:
 * @iter: an initialized SPRpcIter
 * @key: a location for the key, or NULL
 * @value: the location for the value, or NULL
 * @returns: FALSE if the end of struct has been reached
 *
 * Advances iter and retrieves the key and/or value that are now pointed to
 * as a result of this advancement. If FALSE is returned, key and value are
 * not set, and the iterator becomes invalid.
 */
gboolean
sp_rpc_value_struct_iter_next (SPRpcIter * iter, gchar ** key,
			       SPRpcValue ** value)
{
  g_return_val_if_fail (iter != NULL, FALSE);

  return g_hash_table_iter_next (iter, (gpointer *) key, (gpointer *) value);
}

/**
 * sp_rpc_value_struct_iter_remove:
 * @iter: an initialized SPRpcIter
 *
 * Removes the key/value pair currently pointed to by the iterator from its
 * associated SPRpcValue. Can only be called after
 * sp_rpc_value_struct_iter_next() returned TRUE, and cannot be called more
 * than once for the same key/value pair. 
 *
 * The count reference of the removed value will be decreased.
 */
void
sp_rpc_value_struct_iter_remove (SPRpcIter * iter)
{
  g_return_if_fail (iter != NULL);

  g_hash_table_iter_remove (iter);
}

/* EOF */
