/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <sp_rpc.h>

#define DEFAULT_TIMEOUT		5

typedef struct
{
  GIOChannel *io;

  SPRpc *rpc;

  GMainLoop *loop;
  GSource *source;

  gint timeout;

  GList *responses;
} data_t;

typedef struct
{
  SPRpcResponseID id;
  GTimeVal tv;
} response_t;

/* Commands and prototypes: */
static void test_glib_func_list (GList * list, data_t * data);
static void test_glib_func_call (GList * list, data_t * data);
static void test_glib_func_params (GList * list, data_t * data);
static void test_glib_func_exists (GList * list, data_t * data);
static void test_glib_func_getTimer (GList * list, data_t * data);
static void test_glib_func_setTimer (GList * list, data_t * data);
static void test_glib_func_quit (GList * list, data_t * data);
static void test_glib_func_help (GList * list, data_t * data);

struct cmd_t
{
  gchar *cmd;
  gint numb_args;
  void (*function) (GList * args, data_t * data);
} cmds[] =
{
  {
  "list", 0, test_glib_func_list},
  {
  "call", -1, test_glib_func_call},
  {
  "params", 1, test_glib_func_params},
  {
  "exists", 1, test_glib_func_exists},
  {
  "getTimer", 0, test_glib_func_getTimer},
  {
  "setTimer", 1, test_glib_func_setTimer},
  {
  "quit", 0, test_glib_func_quit},
  {
  "help", 0, test_glib_func_help},
  {
  NULL}
};

/* Response functions: */
static void
test_glib_response (data_t * data, SPRpcResponseID id)
{
  response_t *r;

  r = g_malloc0 (sizeof (response_t));

  r->id = id;
  g_source_get_current_time (data->source, &r->tv);

  data->responses = g_list_append (data->responses, r);
  fprintf (stdout, "Added Response %d.\n", id);
}

static void test_glib_func_response_value (SPRpcValue * value, gint indent);

static void
test_glib_func_response (SPRpc * rpc, SPRpcResponseID id,
			 SPRpcParams * params, data_t * data)
{
  SPRpcIter iter;
  SPRpcValue *value;
  gchar *name;

  if (!params || sp_rpc_params_is_empty (params) == TRUE)
    {
      fprintf (stdout, "Response %d: no params\n", id);
      return;
    }

  fprintf (stdout, "Response %d:\n", id);

  sp_rpc_params_iter_init (params, &iter);
  while (sp_rpc_params_iter_next (&iter, &name, &value) == TRUE)
    {
      fprintf (stdout, "Param: %s\n", name);
      fprintf (stdout, "Value: ");

      test_glib_func_response_value (value, 7);
    }
}

static void
test_glib_func_response_value_indent (gint i)
{
  for (; i >= 0; i--)
    fprintf (stdout, " ");
}

static void
test_glib_func_response_value (SPRpcValue * value, gint indent)
{
  switch (sp_rpc_value_get_type (value))
    {
    case SP_RPC_TYPE_NULL:
      fprintf (stdout, "null\n");
      break;

    case SP_RPC_TYPE_NUMBER:
      fprintf (stdout, "number(%f)\n", sp_rpc_value_get_number (value));
      break;

    case SP_RPC_TYPE_BOOLEAN:
      fprintf (stdout, "boolean(%s)\n",
	       sp_rpc_value_get_boolean (value) ? "true" : "false");
      break;

    case SP_RPC_TYPE_STRING:
      {
	gchar *s = sp_rpc_value_get_string (value);
	fprintf (stdout, "string(%s)\n", s);
	g_free (s);
      }

      break;

    case SP_RPC_TYPE_ARRAY:
      {
	gint i, len;
	gboolean first = TRUE;

	fprintf (stdout, "array( ");

	for (i = 0, len = sp_rpc_value_array_length (value); i < len; i++)
	  {
	    SPRpcValue *v;

	    if (first == FALSE)
	      test_glib_func_response_value_indent (indent + 7);
	    else
	      first = FALSE;

	    v = sp_rpc_value_array_get (value, i);
	    test_glib_func_response_value (v, indent + 7);
	  }

	test_glib_func_response_value_indent (indent);
	fprintf (stdout, ")\n");
      }
      break;

    case SP_RPC_TYPE_STRUCT:
      {
	SPRpcIter iter;
	SPRpcValue *v;
	gchar *k;

	gboolean first = TRUE;

	fprintf (stdout, "struct( ");

	sp_rpc_value_struct_iter_init (value, &iter);
	while (sp_rpc_value_struct_iter_next (&iter, &k, &v) == TRUE)
	  {
	    if (first == FALSE)
	      test_glib_func_response_value_indent (indent + 7);
	    else
	      first = FALSE;

	    fprintf (stdout, "Name: %s\n", k);

	    test_glib_func_response_value_indent (indent + 7);
	    fprintf (stdout, "Value: ");

	    test_glib_func_response_value (v, indent + 15);
	  }

	test_glib_func_response_value_indent (indent);
	fprintf (stdout, ")\n");
      }
      break;
    }
}

/* Commands: */
static void
test_glib_func_list (GList * list, data_t * data)
{
  SPRpcResponseID id;

  sp_rpc_call_list (data->rpc, &id, (SPRpcResponse) test_glib_func_response,
		    data);
  test_glib_response (data, id);
}

static void
test_glib_func_call (GList * list, data_t * data)
{
  SPRpcParams *params;
  SPRpcResponseID id;
  gchar *method;

  if (!list)
    {
      fprintf (stderr,
	       "The call command requires the method as first argument.\n");
      return;
    }

  method = list->data;

  params = sp_rpc_params_new ();

  if (list->next)
    {
      for (list = list->next; list && list->next; list = list->next->next)
	{
	  SPRpcValue *v = sp_rpc_value_new ();

	  sp_rpc_value_set_type (v, SP_RPC_TYPE_STRING);
	  sp_rpc_value_set_string (v, list->next->data);

	  sp_rpc_params_add (params, list->data, v);
	  sp_rpc_value_unref (v);
	}

      if (list)
	{
	  fprintf (stderr,
		   "The params arguments are wrong. Please, insert 'name value name value...'\n");
	  sp_rpc_params_unref (params);
	  return;
	}
    }

  sp_rpc_call (data->rpc, method, params, SP_RPC_REQ_TYPE_CALL, &id,
	       (SPRpcResponse) test_glib_func_response, data);
  test_glib_response (data, id);

  sp_rpc_params_unref (params);
}

static void
test_glib_func_params (GList * list, data_t * data)
{
  SPRpcResponseID id;
  gchar *method;

  method = list->data;

  sp_rpc_call (data->rpc, method, NULL, SP_RPC_REQ_TYPE_GET_PARAMS, &id,
	       (SPRpcResponse) test_glib_func_response, data);
  test_glib_response (data, id);
}

static void
test_glib_func_exists (GList * list, data_t * data)
{
  SPRpcResponseID id;
  gchar *method;

  method = list->data;

  sp_rpc_call (data->rpc, method, NULL, SP_RPC_REQ_TYPE_EXISTS, &id,
	       (SPRpcResponse) test_glib_func_response, data);
  test_glib_response (data, id);
}

static void
test_glib_func_getTimer (GList * list, data_t * data)
{
  printf ("%d\n", data->timeout);
}

static void
test_glib_func_setTimer (GList * list, data_t * data)
{
  gint timeout = atoi (list->data);

  if (!timeout)
    fprintf (stderr, "The timer `%s' is not valid.\n", (gchar *) list->data);
  else
    data->timeout = timeout;
}

static void
test_glib_func_quit (GList * list, data_t * data)
{
  g_main_loop_quit (data->loop);
}

static void
test_glib_func_help (GList * list, data_t * data)
{
  gint i;

  fprintf (stdout, "\n");

  for (i = 0; cmds[i].cmd; i++)
    fprintf (stdout, "%s\n", cmds[i].cmd);

  fprintf (stdout, "\n");
}

/* Other stuffs: */
static void
test_glib_function_date (SPRpc * rpc, SPRpcMethodID id, SPRpcParams * params,
			 SPRpcParams * return_params, gpointer user_data)
{
  SPRpcValue *value;
  gchar tmp[1024];
  time_t t;

  t = time (NULL);
  snprintf (tmp, sizeof (tmp), ctime (&t));

  value = sp_rpc_value_new ();
  sp_rpc_value_set_type (value, SP_RPC_TYPE_STRING);
  sp_rpc_value_set_string (value, tmp);

  sp_rpc_params_add (return_params, "date", value);
  sp_rpc_value_unref (value);
}

static void
test_glib_function_echo (SPRpc * rpc, SPRpcMethodID id, SPRpcParams * params,
			 SPRpcParams * return_params, gpointer user_data)
{
  SPRpcValue *value;
  gchar *str;

  if (!(value = sp_rpc_params_get (params, "what")))
    return;

  if (!(str = sp_rpc_value_get_string (value)))
    return;

  fprintf (stdout, "%s\n", str);
  g_free (str);
}

gboolean
test_glib_error_function (SPRpcGlibSource * source, SPRpcStatus status,
			  data_t * data)
{
  fprintf (stderr, "Error message from spread: ");

  switch (status)
    {
    case SP_RPC_STATUS_DONE:
      fprintf (stderr, "success.\n");
      break;

    case SP_RPC_STATUS_ILLEGAL_SESSION:
      fprintf (stderr, "illegal session.\n");
      break;

    case SP_RPC_STATUS_ILLEGAL_MESSAGE:
      fprintf (stderr, "illegal message.\n");
      break;

    case SP_RPC_STATUS_CONNECTION_CLOSED:
      fprintf (stderr, "connection closed.\n");
      break;

    default:
    case SP_RPC_STATUS_UNKNOWN_ERROR:
      fprintf (stderr, "unknown error.\n");
      break;
    }

  g_main_loop_quit (data->loop);
  return TRUE;
}

static SPRpc *
test_glib_connect (gchar * spread)
{
  SPRpc *rpc;
  mailbox mbox;

  gchar group[MAX_GROUP_NAME];

  if (SP_connect (spread, NULL, 0, 0, &mbox, group) != ACCEPT_SESSION)
    {
      fprintf (stderr, "Error connecting to the Spread Ring\n");
      return NULL;
    }

  SP_join (mbox, "TEST_GROUP");

  rpc = sp_rpc_new ();
  sp_rpc_set_mailbox (rpc, mbox);
  sp_rpc_set_group (rpc, "TEST_GROUP");
  sp_rpc_set_id (rpc, "test_sp_rpc_client");

  return rpc;
}

static void
test_glib_disconnect (SPRpc * rpc)
{
  mailbox mbox;

  mbox = sp_rpc_get_mailbox (rpc);
  sp_rpc_unref (rpc);

  SP_leave (mbox, "TEST_GROUP");
  SP_disconnect (mbox);
}

static void
test_glib_register (data_t * data)
{
  SPRpcParams *params;
  SPRpcValue *value;

  sp_rpc_register (data->rpc, "date", "bla bla bla...", NULL,
		   test_glib_function_date, data);

  params = sp_rpc_params_new ();

  value = sp_rpc_value_new ();
  sp_rpc_value_set_type (value, SP_RPC_TYPE_STRING);
  sp_rpc_params_add (params, "what", value);
  sp_rpc_value_unref (value);

  sp_rpc_register (data->rpc, "echo", "bla bla bla...", params,
		   test_glib_function_echo, data);

  sp_rpc_params_unref (params);
}

static void
test_glib_prompt (void)
{
  fprintf (stdout, "TestGlib> ");
  fflush (stdout);
}

static gboolean
test_glib_io (GIOChannel * io, GIOCondition condition, data_t * data)
{
  GIOStatus status;
  gchar **split;
  gchar *line;
  gsize last;
  gint i;

  gchar *cmd = NULL;
  GList *args = NULL;
  GError *error = NULL;

  status = g_io_channel_read_line (io, &line, NULL, &last, &error);

  switch (status)
    {
    case G_IO_STATUS_ERROR:
      fprintf (stderr, "Error: %s\n", error->message);
      g_error_free (error);

      g_main_loop_quit (data->loop);
      return FALSE;

    case G_IO_STATUS_EOF:
      g_main_loop_quit (data->loop);
      return FALSE;

    case G_IO_STATUS_AGAIN:
      return TRUE;

    case G_IO_STATUS_NORMAL:
      break;
    }

  if (last == 0)
    {
      g_free (line);
      test_glib_prompt ();
      return TRUE;
    }

  line[last] = 0;

  split = g_strsplit (line, " ", -1);

  for (i = 0; split && split[i]; i++)
    {
      if (*split[i])
	{
	  cmd = split[i];
	  break;
	}
    }

  if (!cmd)
    {
      g_free (line);
      g_strfreev (split);
      test_glib_prompt ();
      return TRUE;
    }

  for (i++; split && split[i]; i++)
    if (*split[i])
      args = g_list_append (args, split[i]);

  for (i = 0; cmds[i].cmd; i++)
    if (!strcmp (cmds[i].cmd, cmd))
      break;

  if (!cmds[i].cmd)
    fprintf (stdout, "%s: Command not found\n", cmd);

  else if (cmds[i].numb_args != -1
	   && cmds[i].numb_args != g_list_length (args))
    fprintf (stdout, "The command `%s' requires %d arguments.\n", cmd,
	     cmds[i].numb_args);

  else
    cmds[i].function (args, data);

  if (args)
    g_list_free (args);

  g_strfreev (split);
  g_free (line);

  test_glib_prompt ();
  return TRUE;
}

static gboolean
test_glib_gcollector (data_t * data)
{
  GList *list;
  GTimeVal tv;

  g_source_get_current_time (data->source, &tv);

  list = data->responses;
  while (list)
    {
      response_t *r = list->data;

      if (tv.tv_sec >= r->tv.tv_sec + data->timeout)
	{
	  fprintf (stdout, "\nRemoved Response %d.\n", r->id);
	  sp_rpc_call_remove (data->rpc, r->id);
	  g_free (r);

	  list = list->next;
	  data->responses = g_list_remove (data->responses, r);
	  continue;
	}

      list = list->next;
    }

  return TRUE;
}

int
main (void)
{
  data_t data;

  memset (&data, 0, sizeof (data_t));
  data.timeout = DEFAULT_TIMEOUT;

  /* Connect: */
  if (!(data.rpc = test_glib_connect ("4803")))
    return 1;

  /* Register: */
  test_glib_register (&data);

  g_timeout_add (500, (GSourceFunc) test_glib_gcollector, &data);

  /* Source: */
  data.source = sp_rpc_glib_source (data.rpc);
  g_source_set_callback (data.source, (GSourceFunc) test_glib_error_function,
			 &data, NULL);
  g_source_attach (data.source, NULL);

  /* IO Channel for stdin: */
#ifndef G_OS_WIN32
  data.io = g_io_channel_unix_new (0);
#else
  data.io = g_io_channel_win32_new_fd (0);
#endif

  g_io_add_watch (data.io, G_IO_IN, (GIOFunc) test_glib_io, &data);

  /* Prompt: */
  test_glib_prompt ();

  /* Loop: */
  data.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (data.loop);
  g_main_loop_unref (data.loop);

  /* Removing the IO Channel: */
  g_io_channel_shutdown (data.io, FALSE, NULL);
  g_io_channel_unref (data.io);

  /* Freeing the source: */
  g_source_unref (data.source);

  /* Disconnect: */
  test_glib_disconnect (data.rpc);
  return 0;
}
