/* SPRpc - Copyright (C) 2008 Andrea Marchesini <baku@autistici.org>
 *                            Rocco De Patto    <lesion@autistici.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <sp_rpc.h>

static void test_get_value (SPRpcValue * value);
static void test_get_array (SPRpcValue * value);
static void test_get_struct (SPRpcValue * value);

static void test (gchar * title, gchar * format, ...);

int
main (void)
{
  test ("null", "n");
  test ("number", "d", (gdouble) 123);
  test ("boolean", "b", TRUE);
  test ("string", "s", "this is a string");

  test ("empty array", "()");
  test ("simple array", "(dsbd)", 1.1, "test", TRUE, 0.9);
  test ("sub arrays", "(ds(ds)bd)", 1.1, "test", 2.2, "test 2", TRUE, 0.9);

  test ("empty struct", "{}");
  test ("simple struct", "{s:d}", "test", 3.3);
  test ("normal struct", "{s:d,s:b}", "test", 3.3, "test 2", TRUE);
  test ("complex struct", "{s:(dbs),s:s}", "test", 3.3, TRUE, "aa", "bb",
	"cc");
  test ("sub structs", "{s:(db{s:s}),s:s}", "test", 3.3, TRUE, "aa", "bb",
	"cc", "dd", "ee");
  test ("a real structs", "{s:s,s:{s:s,s:d,s:(sss),s:d,s:b}}", "object",
	"book", "description", "title", "bla bla bla", "n_pages",
	(gdouble) 203, "tags", "romantic", "horror", "drammatic", "year",
	(gdouble) 2008, "published", TRUE);

  return 0;
}

static void
test (gchar * title, gchar * format, ...)
{
  SPRpcValue *value;
  va_list va;

  printf ("Testing %s: %s -> ", title, format);
  va_start (va, format);

  if (!(value = sp_rpc_value_builder_va (format, va)))
    printf ("failed.\n");

  else
    {
      test_get_value (value);
      printf ("\n");

      sp_rpc_value_unref (value);
    }

  va_end (va);
}

static void
test_get_value (SPRpcValue * value)
{
  switch (sp_rpc_value_get_type (value))
    {
    case SP_RPC_TYPE_NULL:
      printf ("null");
      break;

    case SP_RPC_TYPE_NUMBER:
      printf ("%f", sp_rpc_value_get_number (value));
      break;

    case SP_RPC_TYPE_BOOLEAN:
      printf ("%s",
	      sp_rpc_value_get_boolean (value) == TRUE ? "true" : "false");
      break;

    case SP_RPC_TYPE_STRING:
      {
	gchar *tmp = sp_rpc_value_get_string (value);
	printf ("\"%s\"", tmp);
	g_free (tmp);
      }
      break;

    case SP_RPC_TYPE_ARRAY:
      printf ("[");
      test_get_array (value);
      printf ("]");
      break;

    case SP_RPC_TYPE_STRUCT:
      printf ("{");
      test_get_struct (value);
      printf ("}");
      break;
    }
}

static void
test_get_array (SPRpcValue * value)
{
  gint i, numb;

  numb = sp_rpc_value_array_length (value);
  for (i = 0; i < numb; i++)
    {
      SPRpcValue *v = sp_rpc_value_array_get (value, i);
      test_get_value (v);

      if (i < numb - 1)
	printf (",");
    }
}

static void
test_get_struct (SPRpcValue * value)
{
  gchar *name;
  SPRpcValue *v;
  SPRpcIter iter;

  gboolean first = TRUE;

  sp_rpc_value_struct_iter_init (value, &iter);
  while (sp_rpc_value_struct_iter_next (&iter, &name, &v) == TRUE)
    {
      if (first == TRUE)
	first = FALSE;
      else
	printf (",");

      printf ("\"%s\":", name);
      test_get_value (v);
    }
}

/* EOF */
